import asyncio
import logging
from collections import namedtuple
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

import redis
from rq import Queue
from rq.job import Job
from rq.registry import FailedJobRegistry, StartedJobRegistry
from rq_scheduler import Scheduler
from sqlalchemy import delete, select
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload

from labfunctions import defaults as df
from labfunctions import errors

# from labfunctions.workflows.registers import register_history_db
from labfunctions.conf.server_settings import settings
from labfunctions.db.sync import SQL
from labfunctions.errors.runtimes import RuntimeNotFound
from labfunctions.executors import ExecID, docker_exec
from labfunctions.managers import projects_mg, runtimes_mg, workflows_mg
from labfunctions.models import WorkflowModel
from labfunctions.notebooks import create_notebook_ctx
from labfunctions.runtimes.builder import builder_exec

# from labfunctions.notebooks import nb_job_executor
from labfunctions.types import ExecutionNBTask, NBTask, ScheduleData, WorkflowDataWeb
from labfunctions.types.runtimes import BuildCtx
from labfunctions.utils import get_version, run_async

_DEFAULT_SCH_TASK_TO = 60 * 5  # 5 minutes


def prepare_notebook_job(
    session, projectid: str, wfid: str, execid: str
) -> ExecutionNBTask:
    """It prepares the task execution of the notebook"""

    wm = workflows_mg.get_by_prj_and_wfid_sync(session, projectid, wfid)
    if wm and wm.enabled:
        task = NBTask(**wm.nbtask)
        runtime = None
        if task.runtime:
            runtime = runtimes_mg.get_by_rid_sync(session, runtime)
        exec_nb_ctx = create_notebook_ctx(projectid, task, execid, runtime, wfid=wfid)
    return exec_nb_ctx


def scheduler_dispatcher(
    projectid: str, wfid: str, execid=None, redis_obj=None, db=None, is_async=True
) -> Union[Job, None]:
    """
    This is the entrypoint of any workflow or job to be executed by RQ and it
    will be executed by :class:`labfunctions.scheduler.SchedulerExecutor`
    in the RQWorker of the control plane.

    This function receive the projectid and wfid as reference of the task to
    be completed. It will use this references to get information about the job
    to execute. In the future, it could act as a router, or each kind of action
    could have their own dispatcher.

    Because rq-scheduler has some limitations:
       1. Doesn't always follow the same functions firms of RQ
       2. wfid is inmutable.
       3. job params are inmutable.
       4. Slow official releases in pypi

    For those reasons, and beacause in the future RQ-Scheduler could be abandoned,
    this abstraction was created, where the idea is to use the RQ-Scheduler only
    to enqueue works through RQ.

    Also, adopting this strategy, allows to react dinamically to changes
    in the workflow

    Parameters
    ----------
    :param projectid: is the id of project, from ProjectModel
    :type str:
    :param wfid: wfid from WorkflowModel
    :type str:

    :return: a Job instance from RQ a or None if the Job or the project is not found
    :rtype: an Union between Job or None.
    """
    _db = db or SQL(settings.SQL)
    _redis = redis_obj or redis.from_url(settings.RQ_REDIS)
    _q = settings.CONTROL_QUEUE  # control_q
    scheduler = SchedulerExecutor(redis_obj=_redis, qname=_q, is_async=is_async)

    logger = logging.getLogger(__name__)

    Session = _db.sessionmaker()

    with Session() as session:
        try:
            # signed = firm_or_new(execid, "dispatcher")
            if not execid:
                execid = str(ExecID())
            execid = ExecID(execid=execid).firm_with(ExecID.types.dispatcher)
            exec_nb_ctx = prepare_notebook_job(session, projectid, wfid, execid)

            scheduler.enqueue_notebook(exec_nb_ctx, qname=exec_nb_ctx.machine)
            logger.info(f"SCHEDULING {wfid} with {execid}")

        except errors.WorkflowNotFound as e:
            logger.error(e)
        except errors.WorkflowDisabled as e:
            logger.warning(e)
            scheduler.cancel_job(wfid)
        except RuntimeNotFound as e:
            logger.error(e)
            scheduler.cancel_job(wfid)


class SchedulerExecutor:
    """
    It manages the logic to enqueue and dispatch jobs.
    The SchedulerExecutor belongs to the server side, it connects the webserver with
    the workers in the control plane.

    Because their main function wraps RQ and RQ-Scheduler some variables names could be
    confusing. When we talk about jobs we talk about the task executed by RQ or RQ-Scheduler.

    :param redis: A Redis instance
    :param qname: configured by default from settings, it MUST BE consistent between the different control plane components.
    """

    def __init__(self, redis_obj: redis.Redis, qname, is_async=True):
        self.redis = redis_obj
        self.Q = Queue(qname, connection=self.redis, is_async=is_async)
        self.qname = qname
        # on_success=rq_job_ok, on_failure=rq_job_error)
        self.scheduler = Scheduler(queue=self.Q, connection=self.redis)
        self.is_async = is_async

    def dispatcher(self, projectid, wfid, execid=None) -> Job:
        """
        Entrypoint of a task execution. Beacause it is a dispatcher it only needs
        the references of job to execute. It will dispatch a job to
        :func:`scheduler_dispatcher` which will prepare the task to be executed in
        a worker.

        Usually this method is called from a web endpoint.

        Sequence of calls for a workflow execution:
             1- Enqueue in the control plane with :func:`scheduler_dispatcher` (control plane)
             2- prepare context and use :method:`enqueue_notebook_in_docker` to
             fire the task to a new queue for the agent. (control plane)
             3- The agent execute the task (data plane)

        Every time a task is enqueue again the step MUST be moved.
        """
        _id = ExecID(execid)
        final_id = _id.firm_with(ExecID.types.dispatcher)

        j = self.Q.enqueue(
            scheduler_dispatcher,
            projectid,
            wfid,
            str(final_id),
            is_async=self.is_async,
            job_id=final_id,
        )
        return j

    def enqueue_notebook(self, nb_job_ctx: ExecutionNBTask, qname=None) -> Job:
        """
        It executes a :class:`labfunctions.types.core.NBTask`
        in the remote machine with runtime configuration of
        the project for this task


        :param nb_job_ctx: a prepared notebook execution task
        :type nb_job_ctx: labfunctions.types.core.ExecutionNBTask
        :param task: NBTask object
        :param execid: An optional execid
        :return: An RQ Job
        :rtype: rq.job.Job
        """
        qname = qname or f"{nb_job_ctx.cluster}.{nb_job_ctx.machine}"

        Q = Queue(qname, connection=self.redis, is_async=self.is_async)
        job = Q.enqueue(
            docker_exec.docker_exec,
            nb_job_ctx,
            job_id=nb_job_ctx.execid,
            job_timeout=nb_job_ctx.timeout,
        )
        return job

    def enqueue_build(self, build_ctx: BuildCtx) -> Job:
        """
        TODO: in the future a special queue should exists.
        TODO: set default timeout for build tasks,
              or better add a type like BuildTaskOptions
        TODO: design internal, onpremise or external docker registries.
        """

        Q = Queue(settings.BUILD_QUEUE, connection=self.redis)
        job = Q.enqueue(
            builder_exec,
            build_ctx,
            job_id=build_ctx.execid,
            job_timeout=(60 * 60) * 24,
        )

        return job

    async def schedule(self, project_id, wfid, wd: WorkflowDataWeb) -> str:
        """Put in RQ-Scheduler a workflows previously created"""
        # schedule_data = data_dict["schedule"]
        # task = NBTask(**data_dict)
        # task.schedule = ScheduleData(**schedule_data)
        # wfid = await workflows_mg.register(session, project_id, task, update=update)

        if wd.schedule.cron:
            await run_async(self._cron2redis, project_id, wfid, wd)
        else:
            await run_async(self._interval2redis, project_id, wfid, wd)

        return wfid

    def _interval2redis(self, projectid: str, wfid: str, wd: WorkflowDataWeb):
        interval = wd.schedule
        start_in_dt = datetime.utcnow() + timedelta(minutes=interval.start_in_min)

        self.scheduler.schedule(
            id=wfid,
            scheduled_time=start_in_dt,
            func=scheduler_dispatcher,
            args=[projectid, wfid],
            interval=interval.interval,
            repeat=interval.repeat,
            queue_name=self.qname,
            timeout=_DEFAULT_SCH_TASK_TO,
        )

    def _cron2redis(self, projectid: str, wfid: str, wd: WorkflowDataWeb):
        cron = wd.schedule
        self.scheduler.cron(
            cron.cron,
            id=wfid,
            func=scheduler_dispatcher,
            args=[projectid, wfid],
            repeat=cron.repeat,
            queue_name=self.qname,
            timeout=_DEFAULT_SCH_TASK_TO,
        )

    def list_jobs(self):
        """List jobs from Redis"""
        jobs = []
        for x in self.scheduler.get_jobs():
            jobs.append(
                dict(
                    wfid=x.id,
                    func_name=x.func_name,
                    created_at=x.created_at.isoformat(),
                )
            )
        return jobs

    def cancel_all(self):
        """Cancel all from redis"""
        for j in self.scheduler.get_jobs():
            self.scheduler.cancel(j)

    def cancel_job(self, wfid):
        """Cancel job from redis"""
        self.scheduler.cancel(wfid)

    async def cancle_job_async(self, wfid):
        await run_async(self.scheduler.cancel, wfid)

    async def delete_workflow(self, session, projectid, wfid: str):
        """Delete job from redis and db"""

        await run_async(self.cancel_job, wfid)
        await workflows_mg.delete_wf(session, projectid, wfid)



import os
import sys

import click

from labfunctions.conf.server_settings import settings
from labfunctions.control_plane import rqscheduler
from labfunctions.types.agent import AgentConfig
from labfunctions.utils import get_external_ip, get_hostname

hostname = get_hostname()


@click.group(name="agent")
def agentcli():
    """
    Execute agent related actions
    """
    pass


@agentcli.command(name="run")
@click.option("--workers", "-w", default=1, help="How many workers spawn")
@click.option("--redis", "-r", default=settings.RQ_REDIS, help="Redis full dsn")
@click.option(
    "--qnames",
    "-q",
    default="default",
    help="Comma separated list of queues to listen to",
)
@click.option(
    "--cluster",
    "-C",
    default="default",
    help="Cluster name, it will be added as qname",
)
@click.option(
    "--ip-address",
    "-i",
    default=None,
    help="IP address of the host",
)
@click.option(
    "--agent-name",
    "-a",
    default=None,
    help="Agent Name",
)
@click.option("--machine-id", "-m", default=f"localhost/ba/{hostname}")
def runcli(redis, workers, qnames, cluster, ip_address, agent_name, machine_id):
    """Run the agent"""
    # pylint: disable=import-outside-toplevel
    from labfunctions.control_plane import agent

    sys.path.append(settings.BASE_PATH)
    os.environ["NB_AGENT_TOKEN"] = settings.AGENT_TOKEN
    os.environ["NB_AGENT_REFRESH_TOKEN"] = settings.AGENT_REFRESH_TOKEN
    os.environ["NB_WORKFLOW_SERVICE"] = settings.WORKFLOW_SERVICE
    ip_address = ip_address or get_external_ip(settings.DNS_IP_ADDRESS)
    queues = qnames.split(",")

    conf = AgentConfig(
        redis_dsn=redis,
        cluster=cluster,
        qnames=queues,
        ip_address=ip_address,
        machine_id=machine_id,
        heartbeat_ttl=settings.AGENT_HEARTBEAT_TTL,
        heartbeat_check_every=settings.AGENT_HEARTBEAT_CHECK,
        agent_name=agent_name,
        workers_n=workers,
    )

    agent.run(conf)


@click.command(name="scheduler")
@click.option("--redis", "-r", default=settings.RQ_REDIS, help="Redis full dsn")
@click.option(
    "--interval", "-i", default=60, help="How often the scheduler checks for work"
)
@click.option("--log-level", "-L", default="INFO")
def rqschedulercli(redis, interval, log_level):
    """Run RQ scheduler"""
    # pylint: disable=import-outside-toplevel
    rqscheduler.run(redis, interval, log_level)



import asyncio
import os
import tempfile
from pathlib import Path
from typing import Any, AsyncGenerator, Dict, Generator, Union

import aiofiles
from smart_open import open as sopen

from labfunctions.utils import mkdir_p

from .kvspec import AsyncKVSpec, GenericKVSpec, KeyReadError, KeyWriteError


class KVLocal(GenericKVSpec):
    """https://googleapis.dev/python/storage/latest/client.html"""

    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket
        self._root = os.getenv("NB_EXT_KV_LOCAL_ROOT", ".nb_tmp")
        mkdir_p(f"{self._root}/{self._bucket}")

    def uri(self, key):
        return f"{self._root}/{self._bucket}/{key}"

    def put(self, key: str, bdata: bytes):
        # obj = io.BytesIO(bdata)
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            with sopen(uri, "wb") as f:
                f.write(bdata)
        except Exception as e:
            raise KeyWriteError(self._bucket, key, str(e))

    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            with sopen(uri, "wb") as f:
                for chunk in generator:
                    f.write(chunk)
        except Exception as e:
            raise KeyWriteError(self._bucket, key, str(e))

        return True

    def get(self, key: str) -> Union[bytes, None]:
        uri = self.uri(key)
        try:
            with sopen(uri, "rb") as f:
                obj = f.read()
                return obj
        except Exception as e:
            raise KeyReadError(self._bucket, key, str(e))

    def from_file_gen(self, fpath) -> Generator[bytes, None, None]:
        for chunk in sopen(fpath, "rb"):
            yield chunk

    def get_stream(self, key: str) -> Generator[bytes, None, None]:
        uri = self.uri(key)
        try:
            for chunk in sopen(uri, "rb"):
                yield chunk
        except Exception as e:
            raise KeyReadError(self._bucket, key, str(e))


class AsyncKVLocal(AsyncKVSpec):
    """For local usage and testing"""

    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket
        self._root = os.getenv("NB_EXT_KV_LOCAL_ROOT", ".nb_tmp")
        mkdir_p(f"{self._root}/{self._bucket}")

    def uri(self, key):
        return f"{self._root}/{self._bucket}/{key}"

    async def put(self, key: str, bdata: bytes):
        uri = self.uri(key)
        mkdir_p(Path(uri).parent)
        try:
            async with aiofiles.open(uri, mode="wb") as f:
                await f.write(bdata)
        except Exception as e:
            raise KeyWriteError(self._bucket, key, str(e))

    async def put_stream(
        self, key: str, generator: Generator[bytes, None, None]
    ) -> bool:
        uri = self.uri(key)
        mkdir_p((Path(uri).parent).resolve())
        try:
            async with aiofiles.open(uri, mode="wb") as f:
                async for data in generator:
                    await f.write(data)
        except Exception as e:
            raise KeyWriteError(self._bucket, key, str(e))

        return True

    async def get(self, key: str) -> Union[bytes, str, None]:
        uri = self.uri(key)
        try:
            async with aiofiles.open(uri, mode="rb") as f:
                data = await f.read()
                return data
        except Exception as e:
            raise KeyReadError(self._bucket, key, str(e))

    async def get_stream(self, key: str) -> AsyncGenerator[bytes, None]:
        """PEP 0525 for Asynchronous generators"""
        uri = self.uri(key)
        try:
            async with aiofiles.open(uri, mode="rb") as f:
                while True:
                    data = await f.read(1024)
                    if not data:
                        break
                    yield data

        except Exception as e:
            raise KeyReadError(self._bucket, key, str(e))



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

from labfunctions.types.machine import (
    BlockInstance,
    BlockStorage,
    ExecMachineResult,
    ExecutionMachine,
    MachineInstance,
    MachineRequest,
)


class ProviderSpec(ABC):
    @abstractmethod
    def create_machine(self, node: MachineRequest) -> MachineInstance:
        pass

    @abstractmethod
    def destroy_machine(self, node: Union[str, MachineInstance]):
        pass

    @abstractmethod
    def list_machines(
        self, location: Optional[str] = None, tags: Optional[List[str]] = None
    ) -> List[MachineInstance]:
        pass

    @abstractmethod
    def create_volume(self, disk: BlockStorage) -> BlockInstance:
        pass

    @abstractmethod
    def destroy_volume(self, disk: Union[str, BlockStorage]) -> bool:
        pass

    @abstractmethod
    def attach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        pass

    @abstractmethod
    def detach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        pass



# pylint: disable=unused-argument
import pathlib
from datetime import datetime
from typing import List, Optional

from pydantic.error_wrappers import ValidationError
from sanic import Blueprint, Sanic
from sanic.response import json
from sanic_ext import openapi

from labfunctions.conf.server_settings import settings
from labfunctions.defaults import API_VERSION
from labfunctions.errors.generics import WorkflowRegisterError
from labfunctions.executors import ExecID

# from labfunctions.executors.context import ExecID, create_notebook_ctx_ondemand
from labfunctions.managers import projects_mg, runtimes_mg, workflows_mg
from labfunctions.notebooks import create_notebook_ctx
from labfunctions.scheduler import SchedulerExecutor, scheduler_dispatcher
from labfunctions.security.web import protected
from labfunctions.types import (
    NBTask,
    ProjectData,
    ScheduleData,
    WorkflowData,
    WorkflowDataWeb,
    WorkflowsList,
)
from labfunctions.utils import (
    get_query_param,
    parse_page_limit,
    run_async,
    secure_filename,
)

from .utils import get_scheduler

workflows_bp = Blueprint("workflows", url_prefix="workflows", version=API_VERSION)

is_async = True
if settings.DEV_MODE:
    is_async = False


@workflows_bp.get("/<projectid>/notebooks/_files")
@openapi.parameter("projectid", str, "path")
@protected()
def notebooks_list(request, projectid):
    """
    List file workflows
    TODO: implement a way to register notebooks from each project
    """
    # pylint: disable=unused-argument

    # nb_files = list_workflows()
    nb_files = []

    return json(nb_files)


@workflows_bp.post("/<projectid>/notebooks/_run")
@openapi.parameter("projectid", str, "path")
@openapi.body({"application/json": NBTask})
@protected()
async def notebooks_run(request, projectid):
    """
    Run a notebook
    """
    # pylint: disable=unused-argument

    # nb_files = list_workflows()
    session = request.ctx.session
    try:
        task = NBTask(**request.json)
    except ValidationError:
        return json(dict(msg="wrong params"), 400)

    execid = ExecID()
    id_ = execid.firm_with(ExecID.types.web)
    runtime = None
    if task.runtime:
        runtime = await runtimes_mg.get_runtime(
            session, projectid, task.runtime, task.version
        )

    nb_ctx = create_notebook_ctx(projectid, task, execid=id_, runtime=runtime)
    scheduler = get_scheduler(request, is_async=is_async)
    await run_async(scheduler.enqueue_notebook, nb_ctx)

    return json(nb_ctx.dict(), 202)


@workflows_bp.get("/<projectid>")
@openapi.parameter("projectid", str, "path")
@openapi.response(200, WorkflowsList, "Notebook Workflow already exist")
@protected()
async def workflows_list(request, projectid):
    """List workflows registered in the database"""
    # pylint: disable=unused-argument

    session = request.ctx.session
    async with session.begin():
        result = await workflows_mg.get_all(session, projectid)
        data = [r.dict() for r in result]

    return json(dict(rows=data), 200)


@workflows_bp.post("/<projectid>")
@openapi.body({"application/json": WorkflowDataWeb})
@openapi.parameter("projectid", str, "path")
@openapi.response(200, {"msg": str}, "Notebook Workflow already exist")
@openapi.response(201, {"wfid": str}, "Notebook Workflow registered")
@openapi.response(400, {"msg": str}, description="wrong params")
@openapi.response(404, {"msg": str}, description="project not found")
@protected()
async def workflow_create(request, projectid):
    """
    Register a notebook workflow and schedule it
    """
    try:
        wfd = WorkflowDataWeb(**request.json)
    except ValidationError:
        return json(dict(msg="wrong params"), 400)

    session = request.ctx.session
    scheduler = get_scheduler(request, is_async=is_async)

    async with session.begin():
        try:
            wfid = await workflows_mg.register(session, projectid, wfd)
            if wfd.schedule and wfd.enabled:
                await scheduler.schedule(projectid, wfid, wfd)
        except WorkflowRegisterError as e:
            return json(dict(msg="workflow already exist"), status=200)

        return json(dict(wfid=wfid), status=201)


@workflows_bp.put("/<projectid>")
@openapi.body({"application/json": WorkflowDataWeb})
@openapi.parameter("projectid", str, "path")
@openapi.response(200, {"wfid": str}, "Notebook Workflow accepted")
@openapi.response(400, {"msg": str}, description="wrong params")
@openapi.response(503, {"msg": str}, description="Error persiting the job")
@protected()
async def workflow_update(request, projectid):
    """
    Register a notebook workflow and schedule it
    """
    try:
        wfd = WorkflowDataWeb(**request.json)
    except ValidationError:
        return json(dict(msg="wrong params"), 400)

    session = request.ctx.session
    scheduler = get_scheduler(request, is_async=is_async)

    async with session.begin():
        try:
            wfid = await workflows_mg.register(session, projectid, wfd, update=True)
            if wfd.schedule and wfd.enabled:
                await scheduler.schedule(projectid, wfid, wfd)
        except WorkflowRegisterError as e:
            return json(dict(msg=str(e)), status=503)

        return json(dict(wfid=wfid), status=201)


@workflows_bp.delete("/<projectid>/<wfid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("wfid", str, "path")
@protected()
async def workflow_delete(request, projectid, wfid):
    """Delete from db and queue a workflow"""
    # pylint: disable=unused-argument
    session = request.ctx.session
    scheduler = get_scheduler(request, is_async=is_async)
    async with session.begin():
        await scheduler.delete_workflow(session, projectid, wfid)
        await session.commit()

    return json(dict(msg="done"), 200)


@workflows_bp.get("/<projectid>/<wfid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("wfid", str, "path")
@openapi.response(200, WorkflowData)
@openapi.response(404, {"msg": str}, description="Job not found")
@protected()
async def workflow_get(request, projectid, wfid):
    """Get a workflow by projectid"""
    # pylint: disable=unused-argument
    session = request.ctx.session
    async with session.begin():
        wd = await workflows_mg.get_by_wfid_prj(session, projectid, wfid)

    if wd:
        return json(wd.dict(), 200)
    return json(dict(msg="Not found"), 404)


@workflows_bp.post("/<projectid>/queue/<wfid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("wfid", str, "path")
@openapi.response(202, dict(execid=str))
@protected()
async def workflow_enqueue(request, projectid, wfid):
    """Enqueue a worflow"""
    # pylint: disable=unused-argument
    sche = get_scheduler(request, is_async=is_async)
    execid = ExecID()
    signed = execid.firm_with(ExecID.types.web)
    job = await run_async(sche.dispatcher, projectid, wfid, signed)

    return json(dict(execid=job.id), 202)


@workflows_bp.post("/<projectid>/_ctx/<wfid>")
@openapi.parameter("projectid", str, "path")
@openapi.parameter("wfid", str, "path")
@openapi.response(200)
@openapi.response(404, {"msg": str}, description="Job not found")
@protected()
async def workflow_generate_ctx(request, projectid, wfid):
    """Enqueue a workflow on demand"""
    # pylint: disable=unused-argument
    execid = ExecID()
    signed = execid.firm_with(ExecID.types.web)

    session = request.ctx.session
    async with session.begin():
        nb_ctx = await workflows_mg.prepare_notebook_job_async(
            session, projectid, wfid, signed
        )
    # job = await run_async(scheduler.dispatcher, projectid, wfid)
    return json(nb_ctx.dict(), 200)



import logging

from sqlalchemy import create_engine, inspect
from sqlalchemy.dialects.postgresql.base import PGInspector

# from sqlalchemy.dialects.postgresql import ARRAY
# from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from labfunctions.db.common import Base


def drop_everything(engine):
    """(On a live db) drops all foreign key constraints before dropping all tables.
    Workaround for SQLAlchemy not doing DROP ## CASCADE for drop_all()
    (https://github.com/pallets/flask-sqlalchemy/issues/722)
    """
    from sqlalchemy.engine.reflection import Inspector
    from sqlalchemy.schema import (
        DropConstraint,
        DropTable,
        ForeignKeyConstraint,
        MetaData,
        Table,
    )

    con = engine.connect()
    trans = con.begin()
    # inspector = Inspector.from_engine(engine)
    inspector = inspect(engine)

    # We need to re-create a minimal metadata with only the required things to
    # successfully emit drop constraints and tables commands for postgres (based
    # on the actual schema of the running instance)
    meta = MetaData()
    tables = []
    all_fkeys = []

    for table_name in inspector.get_table_names():
        fkeys = []

        for fkey in inspector.get_foreign_keys(table_name):
            if not fkey["name"]:
                continue

            fkeys.append(ForeignKeyConstraint((), (), name=fkey["name"]))

        tables.append(Table(table_name, meta, *fkeys))
        all_fkeys.extend(fkeys)

    for fkey in all_fkeys:
        con.execute(DropConstraint(fkey))

    for table in tables:
        con.execute(DropTable(table))

    trans.commit()


class SQL:
    def __init__(self, sqluri: str, pool_size=20, max_overflow=0, inspector=False):
        """
        :param sqluri: 'postgresql://postgres:secret@localhost:5432/twitter'
        """
        self._uri = sqluri
        if "sqlite" in sqluri.split("://", maxsplit=1)[0]:
            self.engine = create_engine(sqluri)
        else:
            self.engine = create_engine(
                sqluri, pool_size=pool_size, max_overflow=max_overflow
            )

        if inspector:
            self.inspector: PGInspector = inspect(self.engine)
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

    def create_all(self):
        Sess = self.scoped_session()
        with Sess() as session:
            with session.begin():
                Base.metadata.create_all(self.engine)

    def drop_all(self):
        # Sess = self.scoped_session()
        # with Sess() as session:
        #    with session.begin():
        #        Base.metadata.drop_all(self.engine)
        drop_everything(self.engine)

    def list_tables(self):
        return self.inspector.get_tables_names()

    def sessionmaker(self, autoflush=True):
        """Session could be used as context manager
        Autocommit will be deprecated in SQLAlchemy 2.0
        Session.begin() method may be used to explicitly start transactions

        :param autoflush: When True, all query operations will issue a Session.
        flush() call to this Session before proceeding. Flush
        """

        session = sessionmaker(bind=self.engine, autoflush=autoflush, future=True)
        return session

    def scoped_session(self, autoflush=True) -> Session:
        """Scoped session return a Session object that could be used
        with a context manager

        https://docs.sqlalchemy.org/en/14/orm/session_basics.html#opening-and-closing-a-session

        """
        SSession = scoped_session(sessionmaker(autoflush=autoflush, bind=self.engine))

        return SSession



from .authentication import auth_from_settings
from .base import AuthSpec, TokenStoreSpec
from .password import PasswordScript
from .utils import get_delta
from .web import get_auth, get_authenticate, protected, sanic_init_auth



import os
import sys
from pathlib import Path

import click
from rich.console import Console
from rich.table import Table

from labfunctions import client, defaults, secrets
from labfunctions.conf import load_client
from labfunctions.errors.client import ProjectUploadError
from labfunctions.utils import execute_cmd

from .utils import watcher

settings = load_client()

console = Console()


@click.group(name="project")
def projectcli():
    """
    Synchronize project with the server

    """


@projectcli.command()
@click.option(
    "--url-service",
    "-u",
    default=settings.WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--agentname",
    "-a",
    default=None,
    help="Agent Name",
)
@click.argument("action", type=click.Choice(["create", "del", "list", "token"]))
def agent(url_service, from_file, agentname, action):
    """[create, list, del, token]  of an agent"""

    c = client.from_file(from_file, url_service)

    if action == "create":
        res = c.projects_create_agent()
        if res:
            console.print("[bold green] Agent created [/]")
        else:
            console.print("[bold yellow] Agent creation fail [/]")

    elif action == "token":
        res = c.projects_agent_token(agentname)
        if res:
            console.print(f"[bold magenta]agent name:[/] {res.agent_name}")
            console.print(f"[bold magenta]access token:[/] {res.creds.access_token}")
            console.print(f"[bold magenta]refresh token:[/] {res.creds.refresh_token}")
        else:
            console.print("[bold yellow] Agent creds fail [/]")

    elif action == "list":
        res = c.projects_agent_list()
        table = Table(title="Agents")
        table.add_column("name", style="cyan", justify="center")
        for a in res:
            table.add_row(a)
        console.print(table)

    elif action == "del":
        if not agentname:
            console.print("[red bold]A agent name must be provided[/]")
            sys.exit(-1)
        res = c.project_agent_delete(agentname)
        if res:
            console.print("[bold green]Agent deleted[/]")
        else:
            console.print("[bold red]Something went wrong[/]")


@projectcli.command(name="list")
@click.option(
    "--url-service",
    "-u",
    default=settings.WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
def listcli(url_service, from_file):
    """List projects"""
    c = client.from_file(from_file, url_service)

    projects = c.projects_list()

    table = Table(title="Projects")
    table.add_column("id", style="cyan", justify="center")
    table.add_column("name", style="cyan", justify="center")
    table.add_column("desc", style="yellow", justify="center")
    for p in projects:
        table.add_row(p.projectid, p.name, p.description)
    console.print(table)


@projectcli.command()
@click.option(
    "--url-service",
    "-u",
    default=settings.WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
def info(url_service, from_file):
    """Project's summary"""
    c = client.from_file(url_service=url_service)
    c.info()



from typing import Dict, List, Set


def scope2dict(scopes: List[str]) -> Dict[str, Set]:
    """Given a list of scopes like ["admin:write", "user:read"]
    it will returns a dictionary where the namespace is the key and the actions
    turns into a set"""
    permissions: Dict[str, Set] = {}
    for s in scopes:
        try:
            ns, action = s.split(":", maxsplit=1)
        except ValueError:
            ns = s
            action = "any"
        ns = ns if ns else "any"
        actions = {a for a in action.split(":")}
        permissions[ns] = actions
    return permissions


def validate(
    scopes: List[str],
    user_scopes: List[str],
    require_all=True,
    require_all_actions=True,
) -> bool:
    """
    from sanic_jwt
    the idea is to provide a scoped access to different resources
    """
    user_perms = scope2dict(user_scopes)
    required = scope2dict(scopes)
    names = {k for k in user_perms.keys()}
    required_names = {k for k in required.keys()}
    intersection = required_names.intersection(names)
    if require_all:
        if len(required_names) != len(intersection) and "any" not in required.keys():
            return False
    elif intersection == 0 and "any" not in required.keys():
        return False

    _match = 0
    for ns in intersection:
        actions_matched = required[ns].intersection(user_perms[ns])
        if len(actions_matched) > 0 or "any" in required[ns]:
            _match += 1
    if _match == 0:
        return False
    return True



from typing import Any, AsyncGenerator, Dict, Generator, Union

import httpx

from .kvspec import AsyncKVSpec, GenericKVSpec


class KVFiles(GenericKVSpec):
    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket

    @property
    def url(self):
        return f"{self._opts['url']}/{self._bucket}"

    def put(self, key: str, bdata: bytes):
        ts = self._opts.get("timeout", 60)
        with httpx.Client(timeout=ts) as client:
            r = client.put(f"{self.url}/{key}", content=bdata)
        if r.status_code == 201:
            return True
        return False

    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        ts = self._opts.get("timeout", 60)
        with httpx.Client(timeout=ts) as client:
            r = client.put(f"{self.url}/{key}", content=generator())
        if r.status_code == 201:
            return True
        return False

    def get(self, key: str) -> Union[bytes, None]:
        with httpx.Client() as client:
            r = client.get(f"{self.url}/{key}")
        if r.status_code == 200:
            return r.content
        return None

    def get_stream(self, key: str) -> Generator[bytes, None, None]:

        with httpx.Client() as client:
            with client.stream("GET", f"{self.url}/{key}") as r:
                for raw in r.iter_raw():
                    yield raw


class AsyncKVFiles(AsyncKVSpec):
    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket

    @property
    def url(self):
        return f"{self._opts['url']}/{self._bucket}"

    async def put(self, key: str, bdata: bytes):
        async with httpx.AsyncClient() as client:
            r = await client.put(f"{self.url}/{key}", content=bdata)
            if r.status_code == 201:
                return True
        return False

    async def put_stream(
        self, key: str, generator: Generator[bytes, None, None]
    ) -> bool:
        ts = self._opts.get("timeout", 60)
        async with httpx.AsyncClient(timeout=ts) as client:
            r = await client.put(f"{self.url}/{key}", content=generator)
            if r.status_code == 201:
                return True
        return False

    async def get(self, key: str) -> Union[bytes, None]:
        ts = self._opts.get("timeout", 60)
        async with httpx.AsyncClient(timeout=ts) as client:
            r = await client.get(f"{self.url}/{key}")
            return r.content

    async def get_stream(self, key: str) -> AsyncGenerator[bytes, None]:
        u = f"{self.url}/{key}"
        ts = self._opts.get("timeout", 60)
        async with httpx.AsyncClient(timeout=ts) as client:
            async with client.stream("GET", u) as r:
                async for chunk in r.aiter_bytes():
                    yield chunk



import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import httpx
from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm, Prompt

from labfunctions import client, defaults, runtimes
from labfunctions.client import from_file
from labfunctions.client.diskclient import DiskClient
from labfunctions.client.nbclient import NBClient
from labfunctions.client.state import WorkflowsState
from labfunctions.conf import load_client
from labfunctions.conf.jtemplates import get_package_dir, render_to_file
from labfunctions.hashes import generate_random
from labfunctions.types import (
    ClientSettings,
    NBTask,
    ProjectData,
    ScheduleData,
    WorkflowDataWeb,
)
from labfunctions.types.docker import DockerfileImage
from labfunctions.types.projects import ProjectCreated
from labfunctions.types.runtimes import RuntimeSpec
from labfunctions.utils import get_parent_folder, get_version, mkdir_p, normalize_name

console = Console()

DIRECTORIES = [
    "nb_app",
    "data",
    "models",
    "outputs",
    defaults.NOTEBOOKS_DIR,
]

PROJECT_FILES = [
    {"tpl": "gitignore", "dst": ".gitignore"},
    {"tpl": "dockerignore", "dst": ".dockerignore"},
    {"tpl": "welcome.ipynb.j2", "dst": "notebooks/welcome.ipynb"},
    {"tpl": "Makefile", "dst": "Makefile"},
    {"tpl": "local.nbvars.j2", "dst": "local.nbvars"},
]


def _example_task() -> NBTask:
    t = NBTask(
        nb_name="welcome",
        description="An example of how to configure a specific workflow",
        params=dict(TIMEOUT=5),
    )
    return t


def _example_workflow() -> WorkflowDataWeb:
    wd = WorkflowDataWeb(
        alias="a_workflow_example",
        nbtask=_example_task(),
        enabled=False,
        # schedule=ScheduleData(
        #    repeat=1,
        #    interval=10,
        # ),
    )
    return wd


def _empty_file(filename):
    with open(filename, "w", encoding="utf-8") as f:
        pass
    return True


def ask_project_name() -> str:
    parent = get_parent_folder()
    _default = normalize_name(parent)
    project_name = Prompt.ask(
        f"Write a name for this project, [yellow]please, avoid spaces and capital "
        "letters[/yellow]: ",
        default=_default,
    )
    name = normalize_name(project_name)
    console.print(
        f"The final name for the project is: [bold magenta]{name}[/bold magenta]"
    )
    return name


def init_nb_app(root, projectid, project_name, url_service=None):
    # _pkg_dir = get_package_dir("labfunctions")
    p = root / "nb_app"

    workflow_service = url_service or "http://localhost:8000"

    mkdir_p(root / "nb_app")
    _empty_file(root / "nb_app" / "__init__.py")
    render_to_file(
        "client_settings.py.j2",
        str((p / "settings.py").resolve()),
        data={
            "projectid": projectid,
            "project_name": project_name,
            "workflow_service": url_service,
        },
    )


def _default_runtime(root):
    version = get_version()
    default_cpu = f"{defaults.DOCKERFILE_IMAGE}:{version}"
    default_gpu = f"{defaults.DOCKERFILE_IMAGE_GPU}:{version}"
    render_to_file(
        "runtimes.yaml",
        str((root / f"runtimes.yaml").resolve()),
        data={"docker_cpu": default_cpu, "docker_gpu": default_gpu},
    )


def init_project_files(root, files):
    _default_runtime(root)
    for f in files:
        render_to_file(
            f["tpl"],
            str((root / f["dst"]).resolve()),
        )
    runtime = runtimes.get_spec_from_file("default")
    runtimes.generate_dockerfile(root, runtime)


def create_folders(root, folders: List[str]):
    for f in folders:
        mkdir_p(root / f)


def workflow_state_init(root, name, projectid=None) -> WorkflowsState:

    wd = _example_workflow()
    wd_dict = {wd.alias: wd}
    # runtime = default_runtime()

    if not projectid:
        projectid = generate_random(
            defaults.PROJECTID_MIN_LEN, alphabet=defaults.PROJECT_ID_ALPHABET
        )

    pd = ProjectData(name=name, projectid=projectid)
    wf_state = WorkflowsState(pd, workflows=wd_dict, version="0.2.0")
    wf_state.write(root / "workflows.yaml")
    return wf_state


def create_on_the_server(
    root, dc: DiskClient, state: WorkflowsState
) -> Union[ProjectCreated, None]:
    rsp = dc.projects_create(
        state.project_name,
        desc=state.project.description,
        repository=state.project.repository,
        store_key=True,
    )
    if rsp:
        dc.state.projectid = rsp.pd.projectid
        # valid_agent = dc.projects_create_agent()
        # agent_creds = dc.projects_agent_token(rsp.pd.agent)
        # with open(f"{dc.working_area}/{defaults.CLIENT_AGENT_CREDS_FILE}", "w") as f:
        #    f.write(agent_creds.json())
        # with open(f"{root}/local.nbvars", "a") as f:
        #    f.write(f"AGENT_TOKEN={agent_creds.access_token}\n")
        #    f.write(f"AGENT_REFRESH_TOKEN={agent_creds.refresh_token}\n")
        return rsp
    return None


def verify_pre_existent(root) -> bool:
    # exist = (root / "local.nbvars").is_file()
    exist = False
    nb_tmp = (root / ".nb_tmp").is_dir()
    wf_file = (root / "workflows.yaml").resolve().is_file()
    if exist or nb_tmp or wf_file:
        return True
    return False


def final_words(project_name, agent_name=None):
    p = Panel.fit(
        "[bold magenta]:smile_cat: Congrats!!!"
        f" Project [cyan]{project_name}[/cyan] created[/bold magenta]",
        border_style="red",
    )
    if agent_name:
        console.print(
            f"Agent for this project was created as: [bold magenta]{agent_name}[/]"
        )
    console.print(p)

    console.print("\n [bold underline magenta]Next steps:[/]")
    console.print("\n\t1. init a git repository")
    console.print("\t2. create a notebook inside of the notebook folder")
    console.print("\t3. generate a workflow for that notebook")
    console.print("\t4. and finally publish your work\n")

    console.print(
        " [bold magenta]To test if everything is working "
        " you can run the following command:[/]\n"
    )
    console.print("\t[bold] nb exec notebook welcome --dev -p TIMEOUT=5[/]\n")


def i_should_create(root):
    ask = True
    if verify_pre_existent(root):
        ask = Confirm.ask(
            "[yellow]It seems that a project already exist, "
            "do you want to continue?[/yellow]",
            default=False,
        )
    return ask


def init_automatic(
    base_path: str,
    project_name: str,
    url_service: str,
    settings: Optional[ClientSettings] = None,
) -> WorkflowsState:
    root = Path(base_path)
    settings = settings or load_client()

    create_folders(root, DIRECTORIES)
    init_project_files(root, PROJECT_FILES)

    state = workflow_state_init(root, project_name)
    init_nb_app(root, state.projectid, project_name, url_service)
    return state


def refresh_project(root, pc: ProjectCreated, url_service: str):
    state = workflow_state_init(root, pc.pd.name, pc.pd.projectid)
    init_nb_app(root, state.projectid, pc.pd.name, url_service)



from pathlib import Path

from labfunctions import defaults
from labfunctions.hashes import generate_random
from labfunctions.types import ProjectData
from labfunctions.types.runtimes import BuildCtx, RuntimeData, RuntimeSpec
from labfunctions.utils import secure_filename

from .utils import get_spec_from_file


def make_docker_name(projectid: str, spec: RuntimeSpec) -> str:
    """
    In the form of
    nbworkflows/[projectid]-[runtime_name]
    """
    return f"{defaults.DOCKER_AUTHOR}/{projectid}-{spec.name}"


def local_spec2runtime(projectid: str, spec: RuntimeSpec, version: str) -> RuntimeData:
    rid = f"{projectid}/{spec.name}/{version}"
    docker_name = make_docker_name(projectid, spec)

    rd = RuntimeData(
        runtimeid=rid,
        spec=spec,
        runtime_name=spec.name,
        docker_name=docker_name,
        project_id=projectid,
        version=version,
        registry=spec.registry,
    )
    return rd


def local_runtime_data(
    projectid, spec_name="default", runtimes_file="runtimes.yaml", version="latest"
) -> RuntimeData:
    spec = get_spec_from_file(spec_name, runtimes_file)
    rd = local_spec2runtime(projectid, spec, version)
    return rd


def execid_for_build(size=defaults.EXECID_LEN):
    return f"bld.{generate_random(size)}"


def build_upload_uri(projectid, runtime_name, version) -> str:
    _name = f"{runtime_name}.{version}.zip"
    name = secure_filename(_name)
    root = Path(projectid)

    uri = str(root / defaults.PROJECT_UPLOADS / name)
    return uri


def create_build_ctx(
    projectid: str,
    spec: RuntimeSpec,
    version: str,
    project_store_class: str,
    project_store_bucket: str,
    registry=None,
) -> BuildCtx:
    _id = execid_for_build()
    uri = build_upload_uri(projectid, spec.name, version)
    docker = make_docker_name(projectid, spec)
    dockerfile = f"Dockerfile.{spec.name}"
    zip_name = uri.split("/")[-1]
    return BuildCtx(
        projectid=projectid,
        spec=spec,
        docker_name=docker,
        version=version,
        dockerfile=dockerfile,
        zip_name=zip_name,
        download_zip=uri,
        execid=_id,
        project_store_class=project_store_class,
        project_store_bucket=project_store_bucket,
        registry=registry,
    )



from sanic import Blueprint, Request, Sanic

from labfunctions import defaults
from labfunctions.conf.server_settings import settings
from labfunctions.io.kvspec import AsyncKVSpec
from labfunctions.scheduler import SchedulerExecutor


def get_query_param2(request, key, default_val=None):
    val = request.args.get(key, default_val)
    return val


def parse_page_limit(request, def_pg="1", def_lt="100"):
    strpage = request.args.get("page", [def_pg])
    strlimit = request.args.get("limit", [def_lt])
    page = int(strpage[0])
    limit = int(strlimit[0])

    return page, limit


def get_scheduler(
    request: Request, qname=settings.CONTROL_QUEUE, is_async=True
) -> SchedulerExecutor:
    current_app = Sanic.get_app(request.app.name)
    r = current_app.ctx.rq_redis
    return SchedulerExecutor(r, qname=qname, is_async=is_async)


def get_kvstore(request: Request) -> AsyncKVSpec:
    return Sanic.get_app(request.app.name).ctx.kv_store


async def stream_reader(request: Request):
    """
    It's a wrapper to be used to yield response from a stream
    to another stream.

    it's used with project upload data to stream upload zip directly to
    the fileserver instead of load data in memory.
    """
    while True:
        body = await request.stream.read()
        if body is None:
            break
        yield body



import os

import click

# from labfunctions.io.fileserver import FileFileserver
import httpx
from rich.console import Console
from rich.table import Table

from labfunctions import client
from labfunctions.client import init_script
from labfunctions.conf import load_client
from labfunctions.utils import mkdir_p

console = Console()


@click.group(name="wf")
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=load_client().WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.pass_context
def workflowscli(ctx, from_file, url_service):
    """
    Start the creation of new workflows in your current folder
    """
    ctx.ensure_object(dict)

    ctx.obj["URL"] = url_service
    ctx.obj["WF_FILE"] = from_file


# @workflowscli.command()
# @click.pass_context
# def init(ctx):
#     """initialize a project"""
#     url_service = ctx.obj["URL"]
#     c = client.init(url_service)
#     c.write()


@click.command()
@click.option(
    "--notebook",
    "-n",
    required=True,
    help="Fullpath to the notebook file",
)
@click.option(
    "--alias",
    "-a",
    required=True,
    help="An alias for this workflow",
)
@click.pass_context
def create(ctx, notebook, alias):
    """Creates a workflow definition from a Notebook,
    if the notebook file doesn't exist, it will be created.
    (Changes aren't pushed to the server until a nb wf push
    is executed)."""
    url_service = ctx.obj["URL"]
    from_file = ctx.obj["WF_FILE"]
    c = client.from_file(from_file, url_service=url_service)
    c.create_workflow(notebook, alias)


@workflowscli.command()
@click.option(
    "--update",
    "-u",
    is_flag=True,
    default=False,
    help="Updates workflows when push",
)
@click.pass_context
def push(ctx, update):
    """Push workflows definitions to the server"""
    action = "updated" if update else "created"
    url_service = ctx.obj["URL"]
    from_file = ctx.obj["WF_FILE"]
    c = client.from_file(from_file, url_service=url_service)
    rsp = c.workflows_push(update=update)
    for r in rsp.errors:
        console.print(f"[bold red]{r.alias} failed[/]")
    for r in rsp.created:
        console.print(f"[bold green]{r.alias} {action} with id: {r.wfid}[/]")

    if not rsp.created and not rsp.errors:
        console.print(f"[bold yellow]No changes[/]")


@workflowscli.command(name="list")
@click.pass_context
def list_wf(ctx):
    """List workflows registered in the server"""
    url_service = ctx.obj["URL"]
    from_file = ctx.obj["WF_FILE"]

    c = client.from_file(from_file, url_service=url_service)
    data = c.workflows_list()

    table = Table(title="Workflows definitions")
    table.add_column("alias", style="cyan", no_wrap=True, justify="center")
    table.add_column("wfid", style="cyan", justify="center")
    table.add_column("is_enabled", justify="right")
    for d in data:
        enabled = "[bold green]yes[/]" if d.enabled else "[bold red]no[/]"
        table.add_row(d.alias, d.wfid, enabled)

    console.print(table)


# @workflowscli.command()
# @click.option("--local", "-l", default=False, is_flag=True, help="execute locally")
# @click.option("--dev", "-d", default=False, is_flag=True, help="execute locally")
# @click.option("--wfid", "-W", required=False, help="workflow id to execute")
# @click.option("--notebook", "-n", required=False, help="notebook name to execute")
# @click.pass_context
# def exec(ctx, local, dev, wfid, notebook):
#     """Exec workflows remote or locally"""
#
#     url_service = ctx.obj["URL"]
#     from_file = ctx.obj["WF_FILE"]
#     if not local and not dev:
#         c = client.from_file(from_file, url_service=url_service)
#         rsp = c.workflows_enqueue(wfid)
#         if rsp:
#             click.echo(f"Executed: {rsp} on the server {url_service}")
#         else:
#             click.echo(f"Something went wrong")
#     elif dev:
#         rsp = local_dev_exec(wfid)
#         if rsp:
#             click.echo(f"Wfid: {rsp.wfid} locally executed")
#             click.echo(f"Executionid: {rsp.execid}")
#             status = "OK"
#             if rsp.error:
#                 status = "ERROR"
#             click.echo(f"Status: {status}")
#     elif local:
#         rsp = local_exec_env()
#         if rsp:
#             click.echo(f"WFID: {rsp.wfid} locally executed")
#             click.echo(f"EXECID: {rsp.execid}")
#             status = "OK"
#             if rsp.error:
#                 status = "ERROR"
#             click.echo(f"Status: {status}")


@workflowscli.command()
@click.argument("wfid", required=True)
@click.pass_context
def delete(ctx, wfid):
    """Delete a workflow definition from server"""
    url_service = ctx.obj["URL"]
    from_file = ctx.obj["WF_FILE"]
    c = client.from_file(from_file, url_service=url_service)
    rsp = c.workflows_delete(wfid)
    c.write()
    print(f"Wfid: {wfid}, deleted. Code {rsp}")

    # elif action == "sync":
    #     c = client.from_file(from_file, url_service=url_service)
    #     c.sync_file()
    #     click.echo(f"{from_file} sync")


workflowscli.add_command(push)
workflowscli.add_command(list_wf)
workflowscli.add_command(delete)
workflowscli.add_command(create)
# workflowscli.add_command(exec)



import json
from typing import Dict, List, Optional, Set, Union

import redis
from rq import Queue
from rq.command import send_shutdown_command

from labfunctions.control_plane.worker import NBWorker
from labfunctions.types.agent import AgentNode
from labfunctions.types.machine import MachineInstance


def workers2dict(workers: List[NBWorker]) -> Dict[str, List[NBWorker]]:
    workers_dict: Dict[str, List[NBWorker]] = {}
    for w in workers:
        agt = w.name.split(".")[0]
        if workers_dict.get(agt):
            workers_dict[agt].append(w)
        else:
            workers_dict[agt] = [w]
    return workers_dict


class AgentRegister:

    AGENT_PREFIX = "nb.ag"
    AGENT_LIST = "nb.agents"
    AGENT_CLUSTERS = "nb.clusters"
    AGENT_CLUSTER_PREFIX = "nb.cluster"
    MACHINE_PREFIX = "nb.mch"
    MACHINE_CLUSTERS = "nb.mch.clusters"
    MACHINE_CLUSTER_PREFIX = "nb.mch.cluster"

    def __init__(self, rdb: redis.Redis, cluster: str):
        """
        The AgentRegister keep track of the agent activity:
        mainly: it register Machine data like ip address, name and workers running.

        It uses the same redis connection of RQ.

        :param rdb: redis.Redis
        """
        self.rdb = rdb
        self.cluster = cluster

    def qname(self, qname):
        return f"{self.cluster}.{qname}"

    def register(self, node: AgentNode):
        """
        Register a `AgentNode`
        """
        key = f"{self.AGENT_PREFIX}.{node.name}"
        pipe = self.rdb.pipeline()
        # for q in node.qnames:
        pipe.sadd(self.AGENT_CLUSTERS, node.cluster)
        pipe.sadd(f"{self.AGENT_CLUSTER_PREFIX}.{node.cluster}", node.name)
        pipe.set(key, node.json())
        pipe.sadd(self.AGENT_LIST, node.name)
        pipe.execute()

    def unregister(self, node: AgentNode):
        key = f"{self.AGENT_PREFIX}.{node.name}"
        pipe = self.rdb.pipeline()
        # for q in node.qnames:
        # pipe.srem(f"{self.CLUSTER_PREFIX}.{q}", node.name)
        pipe.srem(f"{self.AGENT_CLUSTER_PREFIX}.{node.cluster}", node.name)
        pipe.delete(key)
        pipe.srem(self.AGENT_LIST, node.name)
        pipe.execute()

    def get(self, name) -> Union[AgentNode, None]:
        """Get at an agent by name"""

        key = f"{self.AGENT_PREFIX}.{name}"
        jdata = self.rdb.get(key)
        if jdata:
            data = json.loads(jdata)
            return AgentNode(**data)
        return None

    def register_machine(self, node: MachineInstance):
        key = f"{self.MACHINE_PREFIX}.{self.cluster}.{node.machine_id}"
        self.rdb.set(key, node.json())

    def unregister_machine(self, machine_id: str):
        key = f"{self.MACHINE_PREFIX}.{self.cluster}.{machine_id}"
        self.rdb.delete(key)

    def get_machine(self, machine_id: str) -> MachineInstance:
        key = f"{self.MACHINE_PREFIX}.{self.cluster}.{machine_id}"
        data = self.rdb.get(key)
        return MachineInstance(**json.loads(data))

    def get_queue(self, qname) -> Queue:
        return Queue(self.qname(qname), connection=self.rdb)

    def remove(self, agent: str):
        """It removes agent by name from redis"""
        key = f"{self.AGENT_PREFIX}.{agent}"
        pipe = self.rdb.pipeline()
        pipe.srem(self.AGENT_LIST, agent)
        pipe.delete(key)
        pipe.execute()

    def list_clusters(self) -> Set[str]:
        return self.rdb.sinter(self.AGENT_CLUSTERS)

    def list_agents(self, from_cluster=None) -> List[str]:
        """Return a list of the keys of all the agent registered"""
        query = self.AGENT_LIST
        if from_cluster:
            query = f"{self.AGENT_CLUSTER_PREFIX}.{from_cluster}"
        rsp = list(self.rdb.sinter(query))

        if rsp and isinstance(rsp[0], str):
            return rsp
        else:
            return [i.decode("utf-8") for i in rsp]

    def list_agents_by_queue(self, qname: str) -> Set[str]:
        """
        Get agents keys belonging to a queue
        """
        data = set()
        workers = self.get_workers_from_q(self.qname(qname))
        for w in workers:
            data.add(w.name.rsplit(".", maxsplit=1)[0])
        return data

    def get_workers_from_q(self, qname) -> List[NBWorker]:
        q = Queue(self.qname(qname), connection=self.rdb)
        workers = NBWorker.all(queue=q)
        return workers

    def idle_agents_from_cluster(
        self, cluster: str, qnames: List[str]
    ) -> Dict[str, float]:
        """It calculates the idle time from each worker, and then choose the
        lowest by agent. The idle time is returned in minutes
        """
        _qnames = [self.qname(q) for q in qnames]

        queues: List[Queue] = [self.get_queue(q) for q in _qnames]
        agents = self.list_agents(cluster)
        idle_agents = self.idle_agents(agents, queues)
        return idle_agents

    def idle_agents(self, agents: List[str], queues: List[Queue]) -> Dict[str, float]:
        """It calculates the idle time from each worker, and then choose the
        lowest by agent. The idle time is returned in minutes
        if for some reason a workers any worker is found in an agent it will fail
        """
        workers = []
        for q in queues:
            _workers = NBWorker.all(queue=q)
            workers.extend(_workers)

        workers_dict = workers2dict(workers)
        idle_by_agent = {}
        for agt in agents:
            _workers = workers_dict[agt]
            _inactive = [w.inactive_time() for w in _workers]
            minimun = sorted(_inactive)[0]
            idle_by_agent[agt] = minimun / 60
        return idle_by_agent

    def kill_workers_from_agent(self, agent_name) -> bool:
        """
        It sends a shutdown command to RQ Workers from an agent
        """
        node = self.get(agent_name)
        try:
            for w in node.workers:
                send_shutdown_command(self.rdb, w)
            return True
        except AttributeError:
            return False

    def kill_workers_from_queue(self, qname: str):

        q = Queue(self.qname(qname), connection=self.rdb)
        workers = NBWorker.all(queue=q)
        for w in workers:
            send_shutdown_command(self.rdb, w.name)



import asyncio
import codecs
import logging
import os
import pickle
import re
import resource
import socket
import subprocess
import sys
import unicodedata
from datetime import datetime
from functools import wraps
from importlib import import_module
from pathlib import Path
from time import time

import toml
import yaml

from labfunctions import defaults
from labfunctions.errors import CommandExecutionException

_formats = {"hours": "%Y%m%d.%H%M%S", "day": "%Y%m%d", "month": "%Y%m"}
_filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
_windows_device_files = (
    "CON",
    "AUX",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "LPT1",
    "LPT2",
    "LPT3",
    "PRN",
    "NUL",
)


def today_string(utc=True, format_="hours"):
    if utc:
        _now = datetime.utcnow().strftime(_formats[format_])
    else:
        _now = datetime.now().strftime(_formats[format_])
    return _now


def test_error():
    import os

    pid = os.getpid()
    print("From pid: ", pid)

    raise TypeError("error from test_error")


def check_port(ip: str, port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((ip, port))
    if result != 0:
        return False
    return True


def get_free_port(bind="127.0.0.1"):
    sock = socket.socket()
    sock.bind((bind, 0))
    return sock.getsockname()[1]


async def run_async(func, *args, **kwargs):
    """Run sync functions from async code"""
    loop = asyncio.get_running_loop()
    rsp = await loop.run_in_executor(None, func, *args, **kwargs)
    return rsp


def run_sync(func, *args, **kwargs):
    """run async functions from sync code"""
    loop = asyncio.get_event_loop()
    rsp = loop.run_until_complete(func(*args, **kwargs))
    return rsp


def get_query_param(request, key, default_val=None):
    val = request.args.get(key, [default_val])
    return val[0]


def parse_page_limit(request, def_pg="1", def_lt="100"):
    strpage = request.args.get("page", [def_pg])
    strlimit = request.args.get("limit", [def_lt])
    page = int(strpage[0])
    limit = int(strlimit[0])

    return page, limit


def mem():
    """from https://stackoverflow.com/questions/32167386/force-garbage-collection-in-python-to-free-memory"""
    print(
        "Memory usage         : % 2.2f MB"
        % round(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.0, 1)
    )


def format_bytes(n: int) -> str:
    """Format bytes as text

    >>> from dask.utils import format_bytes
    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.21 kiB'
    >>> format_bytes(12345678)
    '11.77 MiB'
    >>> format_bytes(1234567890)
    '1.15 GiB'
    >>> format_bytes(1234567890000)
    '1.12 TiB'
    >>> format_bytes(1234567890000000)
    '1.10 PiB'

    For all values < 2**60, the output is always <= 10 characters.
    """
    for prefix, k in (
        ("Pi", 2**50),
        ("Ti", 2**40),
        ("Gi", 2**30),
        ("Mi", 2**20),
        ("ki", 2**10),
    ):
        if n >= k * 0.9:
            return f"{n / k:.2f} {prefix}B"
    return f"{n} B"


def mem_obj(obj):
    return format_bytes(len(pickle.dumps(obj)))


def mem_df(df):
    return format_bytes(sum(df.memory_usage(deep=True).tolist()))


def flatten_list(list_):
    return [item for sublist in list_ for item in sublist]


def Timeit(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        # print('func:%r args:[%r, %r] took: %2.4f sec' %
        #      (f.__name__, args, kw, te-ts))
        e = round(te - ts, 5)
        print(f"func: {f.__name__} took: {e} sec")
        return result

    return wrap


def Memit(f):
    @wraps(f)
    def wrap(*args, **kw):
        mem()
        result = f(*args, **kw)
        # print(f"func: {f.__name__} took: {e} sec")
        mem()
        return result

    return wrap


def secure_filename(filename: str) -> str:
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.
    On windows systems the function also makes sure that the file is not
    named after one of the special device files.
    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename('i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'
    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you abort or
    generate a random filename if the function returned an empty one.
    .. versionadded:: 0.5
    :param filename: the filename to secure
    """
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.encode("ascii", "ignore").decode("ascii")

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip(
        "._"
    )

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if (
        os.name == "nt"
        and filename
        and filename.split(".")[0].upper() in _windows_device_files
    ):
        filename = f"_{filename}"

    return filename


def get_parent_folder():
    """Get only the name of the parent folder
    commonly used to define the project name
    """
    root = Path(os.getcwd())
    return str(root).rsplit("/", maxsplit=1)[-1]


def open_toml(filepath: str):
    with open(filepath, "r") as f:
        tf = f.read()

    tomconf = toml.loads(tf)
    return tomconf


def write_toml(filepath, data):
    with open(filepath, "w", encoding="utf-8") as f:
        _dump = toml.dumps(data)
        f.write(_dump)


def open_yaml(filepath: str):
    with open(filepath, "r") as f:
        data = f.read()
        dict_ = yaml.safe_load(data)

    return dict_


def write_yaml(filepath: str, data, *args, **kwargs):

    with open(filepath, "w") as f:
        dict_ = yaml.dump(data, *args, **kwargs)
        f.write(dict_)


def set_logger(name: str, level: str):
    logger = logging.getLogger(name)
    _level = getattr(logging, level)
    logger.setLevel(_level)
    return logger


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def execute_cmd(cmd) -> str:
    """Wrapper around subprocess"""
    with subprocess.Popen(
        cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:

        out, err = p.communicate()
        if err:
            raise CommandExecutionException(err.decode())

        return out.decode().strip()


def execute_cmd_no_block(cmd: str, check=True):
    """Wrapper around subprocess
    :param cmd: a string with the command to execute
    :param check: if True then it will checks if the command was ok or raise a
    CalledProccessError.

    """
    res = subprocess.Popen(
        cmd.split(), shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    return res


def path_relative(fp):
    """Given a  filepath returns a normalized a path"""
    return str(Path(fp))


def fullpath(fp):
    """Returns a fullpath based on the BASE_PATH env"""
    base_p = os.getenv(defaults.BASE_PATH_ENV)
    return str(Path(f"{base_p}/{fp}").resolve())


def mkdir_p(fp):
    """Make the fullpath
    similar to mkdir -p in unix systems.
    """
    Path(fp).mkdir(parents=True, exist_ok=True)


def parent_folder():
    return str(Path("../").resolve())


def under_virtualenv() -> bool:
    if sys.prefix == sys.base_prefix:
        return False
    return True


def pkg_route() -> str:
    """
    Get the absoute path of the labfunctions package whatever it is installed

    It could be used to reference files inside of the package.

    :return:
    :type str:
    """
    here = os.path.abspath(os.path.dirname(__file__))
    return here


def read_from_relative(rel_path) -> str:
    here = pkg_route()
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path="__version__.py"):
    """It get the version of NB Workflows package, reading it
    from __version__.py file"""
    for line in read_from_relative(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_api_version(rel_path="__version__.py"):
    for line in read_from_relative(rel_path).splitlines():
        if line.startswith("__api_version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def parse_var_line(line):
    """
    This regex works only if spaces are not used
     ^(\w*)=?*(['|"].*?['|"|])$
    """
    k = line.split("=", maxsplit=1)[0].strip()
    v = line.split("=", maxsplit=1)[1].replace('"', "").strip("\n").strip()
    return k, v


def format_seconds(secs) -> str:
    ago = f"{round(secs)} seconds ago"
    if secs > 60 and secs < 3600:
        ago = f"{round(secs / 60)} minutes ago"
    elif secs >= 3600:
        ago = f"{round((secs / 60)/ 60)} hours ago"
    return ago


def binary_file_reader(fp: str, chunk_size=1024):
    """
    File reader generator mostly used for projects upload
    """
    with open(fp, "rb") as f:
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            yield data


def open_publickey(fp) -> str:
    with open(fp, "r") as f:
        data = f.read()
    return data.strip()


def get_external_ip(dns="8.8.8.8"):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect((dns, 80))
    ip = s.getsockname()[0]
    s.close()
    return ip


def get_internal_ip() -> str:
    return socket.gethostbyname(socket.gethostname())


def get_hostname() -> str:
    return socket.gethostname()


def get_class(fullclass_path):
    """get a class or object from a module. The fullclass_path should be passed as:
    package.my_module.MyClass
    """
    module, class_ = fullclass_path.rsplit(".", maxsplit=1)
    mod = import_module(module)
    cls = getattr(mod, class_)
    return cls


def normalize_name(name: str) -> str:
    """evaluates and verify if a string is secure to be stored"""
    evaluate = name.lower()
    evaluate = evaluate.replace(" ", "_")
    evaluate = secure_filename(name)
    return evaluate



from .base import ProviderSpec
from .cluster_file import load_cluster_file
from .context import machine_from_settings
from .control import ClusterControl
from .inventory import Inventory
from .shortcuts import create_cluster_control, get_spec_from_file



from typing import Any, Dict, List

from labfunctions.types.cluster import ClusterFile, ClusterSpec, ScaleIdle, ScaleItems
from labfunctions.utils import open_yaml

STRATEGIES = {"items": ScaleItems, "idle": ScaleIdle}


def load_spec(spec_data: Dict[str, Any]) -> ClusterSpec:
    """it intiliazie a ClusterSpec"""
    c = ClusterSpec(**spec_data)
    strategies = []
    for strategy in c.policy.strategies:
        s = STRATEGIES[strategy["name"]](**strategy)
        strategies.append(s)
    c.policy.strategies = strategies
    return c


def load_cluster_file(yaml_path) -> ClusterFile:
    """It will open a cluster spec file"""
    data = open_yaml(yaml_path)
    clusters = {}
    for k, v in data["clusters"].items():
        spec = load_spec(v)
        clusters[k] = spec
    inventory = data.get("inventory")
    default_cluster = data.get("default_cluster")

    return ClusterFile(
        clusters=clusters, inventory=inventory, default_cluster=default_cluster
    )



import getpass
import json
from datetime import datetime
from pathlib import Path
from typing import Union

import httpx
import jwt

from labfunctions import defaults
from labfunctions.types import NBTask, ScheduleData
from labfunctions.types import TokenCreds as Credentials
from labfunctions.utils import get_parent_folder, secure_filename


def _example_task() -> NBTask:
    t = NBTask(
        nb_name="test_workflow",
        description="An example of how to configure a specific workflow",
        params=dict(TIMEOUT=5),
        schedule=ScheduleData(
            repeat=1,
            interval=10,
        ),
    )
    return t


def store_credentials_disk(creds: Credentials, homedir: Union[Path, str]):
    hd = Path(homedir)
    creds_path = hd / defaults.CLIENT_CREDS_FILE
    with open(creds_path, "w", encoding="utf-8") as f:
        f.write(json.dumps(creds.dict()))
    creds_path.chmod(0o600)


def get_credentials_disk(homedir: Union[Path, str]) -> Union[Credentials, None]:
    hd = Path(homedir)
    creds = hd / defaults.CLIENT_CREDS_FILE
    try:
        if creds.is_file():
            with open(hd / defaults.CLIENT_CREDS_FILE, "r", encoding="utf-8") as f:
                data = f.read()
                data_dict = json.loads(data)

                return Credentials(**data_dict)
        return None
    except PermissionError:
        return None


def del_credentials_disk(homedir: Union[Path, str]):
    hd = Path(homedir)
    creds = hd / defaults.CLIENT_CREDS_FILE
    if creds.is_file():
        creds.unlink()


def store_private_key(key, working_area: Union[Path, str]):
    wa = Path(working_area)
    wa.mkdir(parents=True, exist_ok=True)
    with open(wa / "private_key", "w", encoding="utf-8") as f:
        f.write(key)
    Path(wa / "private_key").chmod(0o600)


def get_private_key(working_area: Union[Path, str]) -> str:
    wa = Path(working_area)
    try:
        with open(Path(wa / "private_key"), "r", encoding="utf-8") as f:
            key = f.read().strip()
            return key
    except FileNotFoundError:
        return None


def validate_credentials_local(token) -> bool:
    """Validates locally against the jwt signature
    If exp is not bigger than now, then is valid.
    """
    n = datetime.utcnow()
    ts_now = int(datetime.timestamp(n))
    tkn_dict = jwt.decode(token, options={"verify_signature": False})

    if ts_now < tkn_dict["exp"]:
        return True
    return False



import redis
from rq_scheduler.scheduler import Scheduler
from rq_scheduler.utils import setup_loghandlers


def run(redis_dsn, interval, log_level):
    connection = redis.from_url(redis_dsn)
    setup_loghandlers(log_level)
    scheduler = Scheduler(connection=connection, interval=interval)
    scheduler.run()



import json
import logging
import os
import tempfile
from io import BytesIO
from pathlib import Path
from typing import Optional
from zipfile import ZipFile

import httpx

import docker
from labfunctions import client, defaults
from labfunctions.commands import DockerCommand
from labfunctions.conf import load_client, load_server
from labfunctions.io.kvspec import GenericKVSpec

# from labfunctions.types.docker import DockerBuildLog, DockerBuildLowLog, DockerPushLog
from labfunctions.types.docker import DockerBuildLog
from labfunctions.types.runtimes import BuildCtx, RuntimeReq


def unzip_runtime(project_zip_file, dst_dir):
    with ZipFile(project_zip_file, "r") as zo:
        zo.extractall(dst_dir)


class BuildTask:
    def __init__(
        self,
        projectid: str,
        kvstore: GenericKVSpec,
        nbclient: Optional[client.nbclient.NBClient] = None,
        workflow_service: Optional[str] = None,
        agent_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ):
        self.client = nbclient or client.agent(
            url_service=workflow_service,
            token=agent_token,
            refresh=refresh_token,
            projectid=projectid,
        )
        self.projectid = projectid
        self.kv = kvstore

    def get_runtime_file(self, full_zip_file_path, download_key_zip):
        with open(full_zip_file_path, "wb") as f:
            for chunk in self.kv.get_stream(download_key_zip):
                f.write(chunk)

    def run(self, ctx: BuildCtx) -> DockerBuildLog:
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_file = f"{tmp_dir}/{ctx.zip_name}"
            self.get_runtime_file(zip_file, ctx.download_zip)
            unzip_runtime(zip_file, tmp_dir)
            docker_tag = ctx.docker_name
            push = False
            if ctx.registry:
                docker_tag = f"{ctx.registry}/{docker_tag}"
                push = True
            # nb_client.events_publish(
            #    ctx.execid, f"Starting build for {docker_tag}", event="log"
            # )
            cmd = DockerCommand()
            logs = cmd.build(
                f"{tmp_dir}/src",
                f"{ctx.dockerfile}",
                tag=docker_tag,
                version=ctx.version,
                push=push,
            )

        return logs

    def register(self, ctx: BuildCtx):
        req = RuntimeReq(
            runtime_name=ctx.spec.name,
            docker_name=ctx.docker_name,
            spec=ctx.spec,
            project_id=ctx.projectid,
            version=ctx.version,
            registry=ctx.registry,
        )

        self.client.runtime_create(req)


def builder_exec(ctx: BuildCtx):
    """
    It will get the bundle file, build the container and register it
    """
    url_service = os.getenv("NB_WORKFLOW_SERVICE")
    access_token = os.getenv("NB_AGENT_TOKEN")
    refresh_token = os.getenv("NB_AGENT_REFRESH_TOKEN")

    kv = GenericKVSpec.create(ctx.project_store_class, ctx.project_store_bucket)
    task = BuildTask(
        ctx.projectid,
        kvstore=kv,
        workflow_service=url_service,
        agent_token=access_token,
        refresh_token=refresh_token,
    )
    rsp = task.run(ctx)
    if not os.getenv("NB_RUN_LOCAL"):
        task.register(ctx)
    return rsp


# def builder_exec(ctx: BuildCtx):
#    """It's in charge of building docker images for projects"""
#
#    settings = load_server()
#
#    nb_client = client.agent(
#        url_service=settings.WORKFLOW_SERVICE,
#        token=settings.AGENT_TOKEN,
#        refresh=settings.AGENT_REFRESH_TOKEN,
#        projectid=ctx.projectid,
#    )
#    project_dir, tmp_dir = prepare_files(ctx, settings)
#
#    # pd = nb_client.projects_get()
#    docker_tag = ctx.docker_name
#    push = False
#    if ctx.registry:
#        docker_tag = f"{ctx.registry}/{docker_tag}"
#        push = True
#
#    nb_client.events_publish(
#        ctx.execid, f"Starting build for {docker_tag}", event="log"
#    )
#    logs = docker_build(
#        f"{tmp_dir}/src",
#        f"{ctx.dockerfile}",
#        tag=docker_tag,
#        version=ctx.version,
#        push=push,
#    )
#    req = RuntimeReq(
#        runtime_name=ctx.spec.name,
#        docker_name=ctx.docker_name,
#        spec=ctx.spec,
#        project_id=ctx.projectid,
#        version=ctx.version,
#        registry=ctx.registry,
#    )
#    nb_client.runtime_create(req)
#
#    nb_client.events_publish(ctx.execid, data="finished", event="result")
#    nb_client.events_publish(ctx.execid, data="exit", event="control")
#
#    return logs



from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from labfunctions import defaults, errors
from labfunctions.executors.execid import ExecID
from labfunctions.types import ExecutionNBTask, ExecutionResult, NBTask
from labfunctions.types.runtimes import RuntimeData
from labfunctions.utils import get_version, today_string

WFID_TMP = "tmp"


def _dummy_wfid():
    wfid = ExecID(size=defaults.WFID_LEN - len(WFID_TMP))
    return f"{WFID_TMP}{wfid}"


def create_dummy_ctx(
    projectid: str, nb_name: str, params: Dict[str, Any] = {}, execid=None
) -> ExecutionNBTask:
    dummy_id = execid or _dummy_wfid()
    task = NBTask(nb_name=nb_name, params=params)

    ctx = create_notebook_ctx(projectid, task, execid=dummy_id)
    return ctx


def prepare_runtime(runtime: Optional[RuntimeData] = None) -> str:
    if not runtime:
        version = get_version()
        _runtime = f"{defaults.DOCKERFILE_IMAGE}:{version}"
    else:
        _runtime = f"{runtime.docker_name}:{runtime.version}"
        if runtime.registry:
            _runtime = f"{runtime.registry}/{runtime}"
    return _runtime


def create_notebook_ctx(
    projectid: str,
    task: NBTask,
    execid: Optional[str] = None,
    runtime: Optional[RuntimeData] = None,
    wfid: Optional[str] = None,
) -> ExecutionNBTask:
    """It creates the execution context of a notebook based on project and workflow data"""
    # root = Path.cwd()
    root = Path(defaults.NOTEBOOKS_DIR)
    today = today_string(format_="day")
    _now = datetime.utcnow().isoformat()

    wfid = wfid or _dummy_wfid()
    execid = execid or str(ExecID())

    _params = deepcopy(task.params)
    _params["WFID"] = wfid
    _params["EXECID"] = execid
    _params["NOW"] = _now

    nb_filename = f"{task.nb_name}.ipynb"

    papermill_input = str(root / nb_filename)

    output_dir = f"{defaults.NB_OUTPUTS}/ok/{today}"
    error_dir = f"{defaults.NB_OUTPUTS}/errors/{today}"

    output_name = f"{wfid}.{task.nb_name}.{execid}.ipynb"
    _runtime = prepare_runtime(runtime)
    machine = task.machine or defaults.MACHINE_TYPE

    return ExecutionNBTask(
        projectid=projectid,
        wfid=wfid,
        execid=execid,
        nb_name=task.nb_name,
        runtime=_runtime,
        machine=task.machine,
        params=_params,
        pm_input=str(papermill_input),
        pm_output=f"{output_dir}/{output_name}",
        output_name=output_name,
        output_dir=output_dir,
        error_dir=error_dir,
        today=today,
        timeout=task.timeout,
        created_at=_now,
        notifications_ok=task.notifications_ok,
        notifications_fail=task.notifications_fail,
    )



import os
import sys
from pathlib import Path

import click
from rich.console import Console
from rich.progress import Progress, SpinnerColumn
from rich.table import Table

from labfunctions import client, defaults, errors, runtimes, secrets
from labfunctions.conf import load_client
from labfunctions.errors.client import ProjectUploadError
from labfunctions.io.kvspec import GenericKVSpec
from labfunctions.runtimes.builder import builder_exec
from labfunctions.runtimes.context import build_upload_uri, create_build_ctx
from labfunctions.types.runtimes import RuntimeSpec
from labfunctions.utils import execute_cmd, open_yaml

from .utils import watcher

settings = load_client()

console = Console()
progress = Progress(
    SpinnerColumn(),
    "[progress.description]{task.description}",
)


@click.group(name="runtimes")
def runtimescli():
    """
    Generates and deploy runtimes

    """


@runtimescli.command()
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=settings.WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
@click.option(
    "--env-file",
    "-E",
    default="local.nbvars",
    help="env file to be encrypted",
)
@click.option(
    "--stash",
    "-S",
    is_flag=True,
    default=False,
    help="untracked files are zipped",
)
@click.option(
    "--current",
    "-C",
    is_flag=True,
    default=False,
    help="It will zip all the files ignoring .gitignore =P",
)
@click.option(
    "--only-bundle",
    "-z",
    is_flag=True,
    default=False,
    help="Only generates the bundle file",
)
@click.option(
    "--watch",
    "-w",
    is_flag=True,
    default=False,
    help="Get events & logs from the executions",
)
@click.option(
    "--local",
    "-L",
    is_flag=True,
    default=False,
    help="Build runtime locally",
)
@click.argument("name", default="default")
def build(
    url_service, from_file, only_bundle, env_file, current, stash, watch, name, local
):
    """Freeze and build a runtime for your proyect into the server"""
    c = client.from_file(from_file, url_service)
    # prepare secrets
    try:
        pv = c.get_private_key()
    except errors.PrivateKeyNotFound:
        console.print("[cyan bold]=> Not private key found for this project[/]")
        pv = None
        # sys.exit(-1)

    # _agent_token = c.projects_agent_token()

    spec = runtimes.get_spec_from_file(name)
    if not spec:
        console.print(f"[red bold](x) Runtime {name} doesn't exists[/]")
        sys.exit(-1)

    console.print(f"=> Bundling runtime [bold magenta]{name}[/]")
    try:
        zfile = runtimes.bundle_project(c.working_area, spec, pv, stash, current)
    except KeyError:
        console.print(
            f"[red bold](x) requirements file missing "
            f"from {spec.container.requirements}[/]"
        )
        sys.exit(-1)
    except TypeError:
        console.print(
            "[red bold](x) There isn't changes in the git repository to perform a "
            "STASH zip file. For untracked files you should add to "
            "the stash the changes, perform: git add .[/]"
        )
        sys.exit(-1)
    except AttributeError:
        console.print(
            "[red bold](x) Bad option: current and stash are different options[/]"
        )
        sys.exit(-1)

    console.print(f"=> Bundle generated in {zfile.filepath}")
    if not only_bundle and not local:
        try:
            c.projects_upload(zfile)
            console.print("[bold green]=> Succesfully uploaded file[/]")
            execid = c.projects_build(spec, zfile.version)
            if not execid:
                console.print("[bold red](x) Error sending build task [/]")
                sys.exit(-1)

            console.print(f"=> Build task sent with execid: [bold magenta]{execid}[/]")
            if watch:
                watcher(c, execid, stats=False)
        except ProjectUploadError:
            console.print("[bold red](x) Error uploading file[/]")
    elif local:
        PROJECTS_STORE_CLASS = "labfunctions.io.kv_local.KVLocal"
        PROJECTS_STORE_BUCKET = "nbworkflows"
        os.environ["NB_AGENT_TOKEN"] = c.creds.access_token
        os.environ["NB_AGENT_REFRESH_TOKEN"] = c.creds.refresh_token
        os.environ["NB_RUN_LOCAL"] = "yes"
        kv = GenericKVSpec.create(PROJECTS_STORE_CLASS, PROJECTS_STORE_BUCKET)
        ctx = create_build_ctx(
            c.projectid,
            spec,
            zfile.version,
            PROJECTS_STORE_CLASS,
            PROJECTS_STORE_BUCKET,
        )
        kv.put_stream(ctx.download_zip, kv.from_file_gen(zfile.filepath))
        with progress:
            task = progress.add_task(
                f" Building docker image for {name}", start=False, total=1
            )
            rs = builder_exec(ctx)
            progress.advance(task)
        # rs = builder_exec(ctx)
        # console.print(rs)
        console.print(f"=> Image: [magenta]{ctx.docker_name}:{ctx.version}[/]")
        console.print(f"=> Version: [magenta]{ctx.version}[/]")
        if ctx.registry:
            console.print(f"=> Registry: [magenta]{ctx.registry}[/]")
        if rs.error:
            console.print(
                f"[red bold](x) Error building "
                f"[magenta]{ctx.docker_name}:{ctx.version}[/][/]"
            )
            console.print(f"[red]{rs.build_log.logs}")
        else:
            console.print("[green bold]=> Succesfully[/]")

    # elif action == "agent-token":
    #    creds = c.projects_agent_token()
    #    click.echo(f"access token (keep private): \n{creds.access_token}")

    # elif action == "recreate":
    #     c.projects_create()


@runtimescli.command(name="list")
@click.option(
    "--from-file",
    "-f",
    default="workflows.yaml",
    help="yaml file with the configuration",
)
@click.option(
    "--url-service",
    "-u",
    default=settings.WORKFLOW_SERVICE,
    help="URL of the NB Workflow Service",
)
def listcli(from_file, url_service):
    """List of runtimes available for this project"""
    c = client.from_file(from_file, url_service)
    runtimes = c.runtimes_get_all()
    table = Table(title="Runtimes for the project")
    # table.add_column("alias", style="cyan", no_wrap=True, justify="center")
    table.add_column("runtime_name", style="cyan", justify="center")
    table.add_column("docker_name", style="cyan", justify="center")
    table.add_column("version", style="cyan", justify="center")
    for runtime in runtimes:
        table.add_row(str(runtime.runtime_name), runtime.docker_name, runtime.version)
    console.print(table)


@runtimescli.command()
@click.option(
    "--from-file",
    "-f",
    default="runtimes.yaml",
    help="yaml file with the runtime configuration",
)
@click.argument("name", default="default")
def generate(from_file, name):
    """Render  Dockerfile.[name] based on runtimes.yaml"""
    root = Path(os.getcwd())
    spec = runtimes.get_spec_from_file(name, from_file)
    runtimes.generate_dockerfile(root, spec)
    console.print(f"[green]Dockerfile generated as Dockerfile.{name}[/]")


# @projectcli.command(name="runtimes")
# @click.pass_context
# def runtimescli(ctx):
#     """List of runtimes available for this project"""
#     url_service = ctx.obj["URL"]
#     from_file = ctx.obj["WF_FILE"]
#     c = client.from_file(from_file, url_service)
#     runtimes = c.runtimes_get_all()
#     table = Table(title="Runtimes for the project")
#     # table.add_column("alias", style="cyan", no_wrap=True, justify="center")
#     table.add_column("id", style="cyan", justify="center")
#     table.add_column("docker_name", style="cyan", justify="center")
#     table.add_column("version", style="cyan", justify="center")
#     for runtime in runtimes:
#         table.add_row(str(runtime.id), runtime.docker_name, runtime.version)
#     console.print(table)
#
#
# @projectcli.command(name="list")
# @click.pass_context
# def listcli(ctx):
#     """List of runtimes available for this project"""
#     url_service = ctx.obj["URL"]
#     from_file = ctx.obj["WF_FILE"]
#     c = client.from_file(from_file, url_service)
#
#     projects = c.projects_list()
#
#     table = Table(title="Projects")
#     table.add_column("id", style="cyan", justify="center")
#     table.add_column("name", style="cyan", justify="center")
#     table.add_column("desc", style="yellow", justify="center")
#     for p in projects:
#         table.add_row(p.projectid, p.name, p.description)
#     console.print(table)
#
#
# @projectcli.command()
# @click.pass_context
# def info(ctx):
#     """Project's summary"""
#     url_service = ctx.obj["URL"]
#     c = client.from_file(url_service=url_service)
#     c.info()



import os
from datetime import datetime
from typing import List

import redis
from loky import get_reusable_executor

from labfunctions.hashes import generate_random
from labfunctions.types.agent import AgentConfig, AgentNode

from .heartbeat import HeartbeatThread
from .register import AgentRegister
from .worker import start_worker


def run(conf: AgentConfig):
    """
    This is the main function which start workers by agent by machine.

    To ack:
    qnames: will be prepended with the name of the cluster, for instance:
    qnames = ["default", "control"]; cluster = "gpu"
    qnames_fit = ["gpu.default", "gpu.control"]

    :param redis_dsn: redis url connection like redis://localhost:6379/0
    :param qnames: a list of queues to listen to
    :param name: a custom name for this worker
    :param ip_address: the ip as worker that will advertise to Redis.
    :param workers_n: how many worker to run
    """

    name = conf.agent_name or conf.machine_id.rsplit("/", maxsplit=1)[1]
    rdb = redis.from_url(conf.redis_dsn, decode_responses=True)

    heart = HeartbeatThread(
        rdb,
        name,
        ttl_secs=conf.heartbeat_ttl,
        check_every_secs=conf.heartbeat_check_every,
    )
    heart.setDaemon(True)
    heart.start()

    workers_names = [f"{name}.{x}" for x in range(conf.workers_n)]
    cluster_queues = [f"{conf.cluster}.{q}" for q in conf.qnames]

    _now = int(datetime.utcnow().timestamp())
    pid = os.getpid()
    node = AgentNode(
        ip_address=conf.ip_address,
        name=name,
        machine_id=conf.machine_id,
        cluster=conf.cluster,
        pid=pid,
        qnames=conf.qnames,
        workers=workers_names,
        birthday=_now,
    )
    ag = AgentRegister(rdb, cluster=conf.cluster)
    ag.register(node)

    if conf.workers_n > 1:
        _executor = get_reusable_executor(max_workers=conf.workers_n, kill_workers=True)
        _results = [
            _executor.submit(
                start_worker, conf.redis_dsn, cluster_queues, conf.ip_address, name_i
            )
            for name_i in workers_names
        ]
    else:
        start_worker(
            conf.redis_dsn,
            cluster_queues,
            name=workers_names[0],
            ip_address=conf.ip_address,
        )
    ag.unregister(node)
    heart.unregister()



import json
import logging
import os
import shlex
import subprocess
import sys
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

import docker
from labfunctions.types.docker import (
    DockerBuildLog,
    DockerBuildLowLog,
    DockerPushLog,
    DockerResources,
    DockerRunResult,
    DockerVolume,
)
from labfunctions.utils import mkdir_p


def shell(
    command: str, check=True, input=None, cwd=None, silent=False, env=None
) -> subprocess.CompletedProcess:
    """
    Runs a provided command, streaming its output to the log files.
    :param command: A command to be executed, as a single string.
    :param check: If true, will throw exception on failure (exit code != 0)
    :param input: Input for the executed command.
    :param cwd: Directory in which to execute the command.
    :param silent: If set to True, the output of command won't be logged or printed.
    :param env: A set of environment variable for the process to use. If None, the current env is inherited.
    :return: CompletedProcess instance - the result of the command execution.
    """
    if not silent:
        log_msg = (
            f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
            f"Executing: {command}" + os.linesep
        )
        print(log_msg)
        print(log_msg)

    proc = subprocess.run(
        shlex.split(command),
        check=check,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        input=input,
        cwd=cwd,
        env=env,
    )

    if not silent:
        print(proc.stderr.decode())
        print(proc.stdout.decode())

    return proc


def docker_low_build(path, dockerfile, tag, rm=False) -> DockerBuildLowLog:
    """It uses the low API of python sdk.
    :param path: path to the Dockerfile
    :param dockerfile: name of the Dockerfile
    :param tag: fullname of the dokcer image to build
    :param rm: remove intermediate build images
    """

    # obj = _open_dockerfile(dockerfile)
    # build(fileobj=obj...
    _client = docker.APIClient(base_url="unix://var/run/docker.sock")
    generator = _client.build(path=path, dockerfile=dockerfile, tag=tag, rm=rm)
    error = False
    log = ""
    while True:
        try:
            output = generator.__next__()
            output = output.decode().strip("\r\n")
            json_output = json.loads(output)
            if "stream" in json_output:
                logging.info(json_output["stream"].strip("\n"))
                log += json_output["stream"]
            elif "errorDetail" in json_output:
                logging.error(json_output["error"])
                log += json_output["error"]
                error = True

        except StopIteration:
            logging.info("Docker image build complete.")
            log += "Docker image build complete.\n"
            break
        except ValueError:
            logging.info("Error parsing output from docker image build: %s" % output)
            log += "Error parsing output from docker image build:{output}\n"
            # raise ValueError(log)
            error = True

    return DockerBuildLowLog(error=error, logs=log)


class DockerCommand:
    __slots__ = "docker"

    def __init__(self, docker_client=None):
        self.docker = docker_client or docker.from_env()

    def _wait_result(
        self, container: docker.models.containers.Container, timeout: int
    ) -> Union[Dict[str, Any], None]:
        result = None
        try:
            result = container.wait(timeout=timeout)
        except Exception:
            pass
        return result

    def run(
        self,
        cmd: str,
        image: str,
        timeout: int = 120,
        env_data: Dict[str, Any] = {},
        remove: bool = True,
        require_gpu: bool = False,
        resources=DockerResources(),
        volumes: List[DockerVolume] = [],
    ) -> DockerRunResult:

        runtime = None
        if require_gpu:
            runtime = "nvidia"

        logs = ""
        status_code = -1
        try:
            container = self.docker.containers.run(
                image,
                cmd,
                runtime=runtime,
                detach=True,
                environment=env_data,
                **resources.dict(),
            )
            result = self._wait_result(container, timeout)
            if not result:
                container.kill()
            else:
                status_code = result["StatusCode"]
            logs = container.logs().decode("utf-8")
            if remove:
                container.remove()
        except docker.errors.ContainerError as e:
            logs = str(e)
            status_code = -2
        except docker.errors.APIError as e:
            logs = str(e)
            status_code = -3
        return DockerRunResult(msg=logs, status=status_code)

    def build(
        self, path: str, dockerfile: str, tag: str, version: str, rm=False, push=False
    ) -> DockerBuildLog:
        """Build docker
        :param path: path to the Dockerfile
        :param dockerfile: name of the Dockerfile
        :param tag: fullname of the dokcer image to build
        :param rm: remove intermediate build images
        :param push: Push docker image to a repository
        """

        error = False
        error_build = False
        error_push = False

        build_log = docker_low_build(path, dockerfile, tag, rm)
        if not build_log.error:
            img = self.docker.images.get(tag)
            img.tag(tag, tag=version)

        error_build = build_log.error

        push_log = None
        if push:
            # push_log = docker_push_image(tag)
            push_log = self.push_image(f"{tag}:{version}")
            error_push = push_log.error

        if error_build or error_push:
            error = True

        return DockerBuildLog(build_log=build_log, push_log=push_log, error=error)

    def push_image(self, tag) -> DockerPushLog:
        """
        Push to docker registry
        :param tag: full name of the docker image to push, it should include
        the registry url
        """

        error = False
        try:
            push_log_str = self.docker.images.push(tag)
        except docker.errors.APIError as e:
            error = True
            push_log_str = str(e)

        return DockerPushLog(logs=push_log_str, error=error)



from .history_client import HistoryClient
from .projects_client import ProjectsClient
from .workflows_client import WorkflowsClient


class NBClient(WorkflowsClient, ProjectsClient, HistoryClient):
    """An staleless client without disk side effects"""

    pass



from collections import namedtuple
from typing import List, NamedTuple

from labfunctions import defaults
from labfunctions.hashes import generate_random


class FirmsTypes(NamedTuple):
    start: str = "0"
    build: str = "bld"
    dispatcher: str = "dsp"
    docker: str = "dck"
    web: str = "web"
    local: str = "loc"
    machine: str = "mch"


class ExecID:

    types = FirmsTypes()

    def __init__(self, execid=None, size=defaults.EXECID_LEN):
        """A general class with manage executions id for different queues in RQ"""
        self._id = execid or generate_random(size=size)

    def firm_by_type(self, firm_type) -> str:
        """it will look at self.types to find the sign"""
        _firm = getattr(self.types, firm_type)
        self.firm_with(_firm)
        return self._id

    def firm_with(self, firm) -> str:
        """it should prepend the word"""
        self._id = f"{firm}.{self._id}"
        return self._id

    def pure(self) -> str:
        pure = self._id
        if "." in self._id:
            pure = self._id.rsplit(".", maxsplit=1)[1]
        return pure

    @property
    def id(self):
        return self._id

    def __str__(self):
        return self._id

    def __repr__(self):
        return self._id



import json
import logging
import os
import shutil
import time
import warnings
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from labfunctions import defaults
from labfunctions.client.diskclient import DiskClient
from labfunctions.client.nbclient import NBClient
from labfunctions.commands import DockerCommand, DockerRunResult
from labfunctions.types import ExecutionNBTask, ExecutionResult, NBTask
from labfunctions.types.runtimes import RuntimeData
from labfunctions.utils import get_version, today_string

from .execid import ExecID

warnings.filterwarnings("ignore", category=DeprecationWarning)


def _prepare_runtime(runtime: Optional[RuntimeData] = None) -> str:
    if not runtime:
        version = get_version()
        _runtime = f"{defaults.DOCKERFILE_IMAGE}:{version}"
    else:
        _runtime = f"{runtime.docker_name}:{runtime.version}"
        if runtime.registry:
            _runtime = f"{runtime.registry}/{runtime}"
    return _runtime


def _simple_retry(func, params, max_retries=3, wait_time=5):
    status = False
    tries = 0
    while status is not True and tries < 3:
        status = func(*params)
        tries += 1
        time.sleep(3)
    return status


class NBTaskExecBase:

    WFID_TMP = "tmp"

    def __init__(self, client: Union[NBClient, DiskClient]):
        self.client = client
        self.logger = logging.getLogger("nbworkf.server")

    @property
    def projectid(self) -> str:
        return self.client.projectid

    def _dummy_wfid(self):
        wfid = ExecID(size=defaults.WFID_LEN - len(self.WFID_TMP))
        return f"{self.WFID_TMP}{wfid}"

    def run(self, ctx: ExecutionNBTask) -> ExecutionResult:
        raise NotImplementedError()

    def register(self, result: ExecutionResult):
        self.client.history_register(result)
        if result.output_name:
            self.client.history_nb_output(result)

    def notificate(self, ctx: ExecutionNBTask, result: ExecutionResult):
        raise NotImplementedError()


class NBTaskDocker(NBTaskExecBase):
    cmd = "nb exec local"

    def build_env(self, data: Dict[str, Any]) -> Dict[str, Any]:
        priv_key = self.client.projects_private_key()
        if not priv_key:
            raise IndexError("No priv key found")

        env = {
            defaults.PRIVKEY_VAR_NAME: priv_key,
            defaults.EXECUTIONTASK_VAR: json.dumps(data),
            "NB_WORKFLOW_SERVICE": self.client._addr,
            defaults.BASE_PATH_ENV: "/app",
        }
        return env

    def run(self, ctx: ExecutionNBTask) -> ExecutionResult:
        _started = time.time()
        env = self.build_env(ctx.dict())
        agent_token = self.client.projects_agent_token(projectid=ctx.projectid)
        env.update(
            {
                "NB_AGENT_TOKEN": agent_token.creds.access_token,
                "NB_AGENT_REFRESH_TOKEN": agent_token.creds.refresh_token,
            }
        )
        cmd = DockerCommand()
        result = cmd.run(
            self.cmd, ctx.runtime, ctx.timeout, env_data=env, require_gpu=ctx
        )
        error = False
        if result.status != 0:
            error = True

        elapsed = round(time.time() - _started)
        return ExecutionResult(
            projectid=ctx.projectid,
            execid=ctx.execid,
            wfid=ctx.wfid,
            cluster=ctx.cluster,
            machine=ctx.machine,
            runtime=ctx.runtime,
            name=ctx.nb_name,
            params=ctx.params,
            input_=ctx.pm_input,
            elapsed_secs=elapsed,
            error=error,
            error_msg=result.msg,
            created_at=ctx.created_at,
        )

    def notificate(self, ctx: ExecutionNBTask, result: ExecutionResult):
        pass


class NBTaskLocal(NBTaskExecBase):
    def run(self, ctx: ExecutionNBTask) -> ExecutionResult:
        import papermill as pm

        _started = time.time()
        _error = False
        Path(ctx.output_dir).mkdir(parents=True, exist_ok=True)
        try:
            pm.execute_notebook(ctx.pm_input, ctx.pm_output, parameters=ctx.params)
        except pm.exceptions.PapermillExecutionError as e:
            self.logger.error(f"jobdid:{ctx.wfid} execid:{ctx.execid} failed {e}")
            _error = True
            self._error_handler(ctx)

        elapsed = time.time() - _started
        return ExecutionResult(
            wfid=ctx.wfid,
            execid=ctx.execid,
            projectid=ctx.projectid,
            name=ctx.nb_name,
            params=ctx.params,
            input_=ctx.pm_input,
            output_dir=ctx.output_dir,
            output_name=ctx.output_name,
            error_dir=ctx.error_dir,
            error=_error,
            elapsed_secs=round(elapsed, 2),
            created_at=ctx.created_at,
        )

    def notificate(self, ctx: ExecutionNBTask, result: ExecutionResult):
        pass

    def _error_handler(self, etask: ExecutionNBTask):
        error_output = f"{etask.error_dir}/{etask.output_name}"
        Path(etask.error_dir).mkdir(parents=True, exist_ok=True)
        shutil.move(etask.pm_output, error_output)



from typing import Dict, List, Optional

from labfunctions.types.machine import BlockStorage, MachineOrm, SSHKey
from labfunctions.utils import Singleton, get_class, open_yaml, pkg_route

from .base import ProviderSpec


class Inventory(metaclass=Singleton):
    def __init__(self, inventory_path: Optional[str] = None):
        path = inventory_path or f"{pkg_route()}/conf/machines.yaml"
        data = open_yaml(path)
        self._inventory_from = path
        self.machines: Dict[str, MachineOrm] = {
            k: MachineOrm(**m) for k, m in data["machines"].items()
        }
        self.volumes: Dict[str, BlockStorage] = {
            k: BlockStorage(**v) for k, v in data["volumes"].items()
        }
        self.providers: Dict[str, str] = data["providers"]

    def reload(self, inventory_path):
        path = inventory_path
        data = open_yaml(path)
        self._inventory_from = path
        self.machines: Dict[str, MachineOrm] = {
            k: MachineOrm(**m) for k, m in data["machines"].items()
        }
        self.volumes: Dict[str, BlockStorage] = {
            k: BlockStorage(**v) for k, v in data["volumes"].items()
        }
        self.providers: Dict[str, str] = data["providers"]

    def machines_by_provider(self, provider: str) -> List[MachineOrm]:
        machines = []
        for k, m in self.machines.items():
            if m.provider == provider:
                machines.append(m)
        return machines

    def get_provider(self, name, *args, **kwargs) -> ProviderSpec:
        prov: ProviderSpec = get_class(self.providers[name])(*args, **kwargs)
        return prov



from typing import Any, Dict, List, Optional, Union

from labfunctions import defaults, errors
from labfunctions.hashes import generate_random
from labfunctions.types import ServerSettings
from labfunctions.types.machine import (
    BlockStorage,
    ExecutionMachine,
    MachineGPU,
    MachineOrm,
    MachineRequest,
    SSHKey,
)
from labfunctions.utils import get_version, open_publickey, open_yaml

from .inventory import Inventory
from .utils import get_local_machine, get_local_volume, ssh_from_settings


def create_machine_ctx(
    machine: MachineOrm,
    qnames: List[str],
    cluster: str,
    agent_homedir=defaults.AGENT_HOMEDIR,
    volumes: List[BlockStorage] = [],
    ssh_key: Optional[SSHKey] = None,
    dynamic_workers=True,
    docker_uid=defaults.DOCKER_UID,
    docker_gid=defaults.DOCKER_GID,
    docker_image=defaults.AGENT_DOCKER_IMG,
    docker_version=None,
) -> ExecutionMachine:
    """
    Its build a machine execution context

    :param machine: A MachineOrm instance
    :param ssh_key: keys for the worker
    :param tags: A list of tags to put to the VM created in the cloud provider
    :param dynamic_workers: if true it will use vcpus in machine_type to allocate workers,
    if false, then only one worker will be allocated.
    """

    execid = f"mch.{generate_random(8)}"
    version = docker_version or get_version()

    _id = generate_random(size=6, alphabet=defaults.NANO_MACHINE_ALPHABET)
    name = f"{machine.name}-{_id}"
    type_ = machine.machine_type
    worker_procs = 1
    if dynamic_workers:
        worker_procs = machine.machine_type.vcpus

    public_key = None
    ssh_user = None
    if ssh_key:
        public_key = open_publickey(ssh_key.public_path)
        ssh_user = ssh_key.user

    gpu = "no"
    if machine.gpu:
        gpu = "yes"

    labels = {"cluster": cluster, "gpu": gpu, "tags": [defaults.CLOUD_TAG]}

    req = MachineRequest(
        name=name,
        ssh_public_cert=public_key,
        ssh_user=ssh_user,
        gpu=machine.gpu,
        image=type_.image,
        volumes=volumes,
        size=type_.size,
        location=machine.location,
        network=type_.network,
        labels=labels,
    )
    ctx = ExecutionMachine(
        execid=execid,
        machine_name=name,
        machine=req,
        provider=machine.provider,
        cluster=cluster,
        qnames=qnames,
        agent_homedir=agent_homedir,
        worker_procs=worker_procs,
        ssh_key=ssh_key,
        docker_uid=docker_uid,
        docker_gid=docker_gid,
        docker_image=docker_image,
        docker_version=version,
    )
    return ctx


def machine_from_settings(
    machine_name: str,
    cluster: str,
    qnames: List[str],
    settings: ServerSettings,
    network: Optional[str] = None,
    location: Optional[str] = None,
    docker_version=None,
    inventory: Optional[Inventory] = None,
) -> ExecutionMachine:
    """It will create a context from settings and the inventory file
    if network or location are set (mainly from the clusterspec) then
    it will will overwrite the definition from inventory.
    """

    inventory = inventory or Inventory()

    m: MachineOrm = inventory.machines[machine_name]
    if network:
        m.machine_type.network = network
    if location:
        m.location = location
    volumes = [inventory.volumes[vol] for vol in m.volumes]

    ssh = ssh_from_settings(settings)
    # agent_env_file = settings.AGENT_ENV_FILE
    ctx = create_machine_ctx(
        m,
        qnames,
        cluster,
        volumes=volumes,
        agent_homedir=settings.AGENT_HOMEDIR,
        ssh_key=ssh,
        docker_version=docker_version,
        docker_uid=settings.DOCKER_UID,
        docker_gid=settings.DOCKER_GID,
    )
    return ctx



from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple, Union

from sqlalchemy import delete, exc
from sqlalchemy import insert as sqlinsert
from sqlalchemy import select

# from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import selectinload

from labfunctions import defaults, secrets
from labfunctions.conf.server_settings import settings
from labfunctions.hashes import generate_random
from labfunctions.managers import users_mg
from labfunctions.models import ProjectModel, UserModel, assoc_projects_users
from labfunctions.types import ProjectData, ProjectReq
from labfunctions.utils import get_parent_folder, normalize_name


def select_project():
    stmt = (
        select(ProjectModel).options(selectinload(ProjectModel.owner))
        # .options(selectinload(ProjectModel.agent))
        .options(selectinload(ProjectModel.users))
    )

    return stmt


def _model2projectdata(obj: ProjectModel) -> ProjectData:
    # agent_name = None
    # if obj.agent:
    #     agent_name = obj.agent.username

    pd = ProjectData(
        name=obj.name,
        projectid=obj.projectid,
        owner=obj.owner.username,
        users=[u.username for u in obj.users],
        # agent=agent_name,
        description=obj.description,
        respository=obj.repository,
    )
    return pd


def _insert(user_id, pq: ProjectReq, projectid):

    table = ProjectModel.__table__
    name = normalize_name(pq.name)
    stmt = sqlinsert(table).values(
        name=name,
        private_key=pq.private_key.encode("utf-8"),
        projectid=projectid,
        description=pq.description,
        repository=pq.repository,
        owner_id=user_id,
    )
    return stmt


def _insert_relation_project(user_id, project_id):
    stmt = sqlinsert(assoc_projects_users).values(
        project_id=project_id, user_id=user_id
    )
    return stmt


def generate_projectid(name=None) -> str:
    return generate_random(
        settings.PROJECTID_LEN, alphabet=defaults.PROJECT_ID_ALPHABET
    )


async def create(session, user_id: int, pq: ProjectReq) -> Union[ProjectData, None]:

    projectid = pq.projectid or generate_projectid()
    stmt = _insert(user_id, pq, projectid)
    project = _insert_relation_project(user_id, projectid)
    # session.add(pm)
    try:
        await session.execute(stmt)
        await session.execute(project)
        agent_user = await create_agent(session, projectid)
        # await session.commit()
        await session.commit()
    except exc.IntegrityError as e:
        # await session.rollback()
        return None

    return ProjectData(
        name=pq.name,
        projectid=projectid,
        description=pq.description,
        repository=pq.repository,
        agent=agent_user.username,
    )


async def assign_project(session, username: str, projectid: str) -> bool:
    """
    Assign a user to a project.
    """
    um = await users_mg.get_user_async(session, username)
    if um:
        # projects = [p.projectid for p in um.projects]
        # if projectid in projects:
        stmt = _insert_relation_project(um.id, projectid)
        await session.execute(stmt)
        return True
    return False


async def create_or_update(session, user_id: int, pq: ProjectReq):
    name = normalize_name(pq.name)
    projectid = generate_projectid()
    pm = ProjectModel(
        name=name,
        private_key=pq.private_key.encode("utf-8"),
        projectid=projectid,
        description=pq.description,
        repository=pq.repository,
        owner_id=user_id,
    )

    # stmt = _create_or_update_stmt(name, user_id, projectid, pq)
    await session.merge(pm)
    # obj = await get_by_name_model(session, name)
    return pm


async def get_by_projectid_model(session, projectid) -> Union[ProjectModel, None]:
    stmt = select_project().where(ProjectModel.projectid == projectid).limit(1)
    r = await session.execute(stmt)
    return r.scalar_one_or_none()


async def get_by_projectid(
    session, projectid, user_id=None
) -> Union[ProjectData, None]:

    stmt = select_project().where(ProjectModel.projectid == projectid)
    if user_id:
        stmt = stmt.where(ProjectModel.owner_id == user_id)
    stmt = stmt.limit(1)
    r = await session.execute(stmt)
    obj: Optional[ProjectModel] = r.scalar_one_or_none()
    if not obj:
        return None
    return _model2projectdata(obj)


async def list_all(session, user_id=None) -> Union[List[ProjectData], None]:
    stmt = select_project()
    if user_id:
        stmt = stmt.where(ProjectModel.owner_id == user_id)
    rows = await session.execute(stmt)

    if rows:
        results = [_model2projectdata(r[0]) for r in rows]
        return results
    return None


# async def list_by_user(session, user_id) -> Union[List[ProjectData], None]:
#     stmt = select_project().where(ProjectModel.owner_id == user_id)
#     rows = await session.execute(stmt)
#     if rows:
#         results = [_model2projectdata(r[0]) for r in rows]
#         return results
#     return None


async def delete_by_projectid(session, projectid):
    """TODO: In the future this could be a task
    or a flag that a async task takes to clean project data
    such as docker images, files and so on.
    """
    stmt = delete(ProjectModel).where(ProjectModel.projectid == projectid)
    await session.execute(stmt)


async def create_agent(
    session,
    projectid=None,
    scopes=defaults.AGENT_SCOPES,
    admin_scopes=None,
    is_admin=False,
):
    prefix = generate_random(size=5)
    if projectid:
        prefix = projectid[:5]
    name = f"{prefix}-{generate_random(defaults.AGENT_LEN)}"
    fullname = f"{defaults.AGENT_USER_PREFIX}{name}"
    if is_admin:
        scopes = defaults.AGENT_ADMIN_SCOPES

    u = UserModel(
        username=fullname,
        scopes=scopes,
        is_agent=True,
    )
    if projectid:
        prj = await get_by_projectid_model(session, projectid)
        u.projects.append(prj)
        prj.updated_at = datetime.utcnow()
        session.add(prj)

    session.add(u)

    return u


async def get_agent(
    session, agentname: str, projectid: Optional[str] = None
) -> Union[UserModel, None]:
    # prj = await get_by_projectid_model(session, projectid)
    stmt = users_mg.join_with_project(agentname)
    res = await session.execute(stmt)
    return res.scalar_one_or_none()


async def get_agent_project(session, projectid) -> Union[UserModel, None]:
    stmt = (
        users_mg.join_by_project()
        .where(UserModel.is_agent.is_(True))
        .where(ProjectModel.projectid == projectid)
        .limit(1)
    )
    res = await session.execute(stmt)
    return res.scalar_one_or_none()


async def get_agent_list(session, projectid) -> List[str]:
    stmt = (
        users_mg.join_by_project()
        .where(UserModel.is_agent.is_(True))
        .where(ProjectModel.projectid == projectid)
    )
    res = await session.execute(stmt)
    return [r[0].username for r in res]


async def delete_agent(session, agentname, projectid: Optional[str] = None) -> bool:
    user = await users_mg.get_user_async(session, agentname, projectid)
    if user:
        await session.delete(user)
        return True
    return False
    # session.add(prj)


def get_private_key_sync(session, project_id) -> Union[str, None]:
    stmt = select(ProjectModel).where(ProjectModel.projectid == project_id).limit(1)
    r = session.execute(stmt)
    obj: Optional[ProjectModel] = r.scalar_one_or_none()
    if obj:
        return obj.private_key.decode("utf-8")
    return None


async def get_private_key(session, project_id) -> Union[str, None]:
    stmt = select(ProjectModel).where(ProjectModel.projectid == project_id).limit(1)
    r = await session.execute(stmt)
    obj: Optional[ProjectModel] = r.scalar_one_or_none()
    if obj:
        return obj.private_key.decode("utf-8")
    return None



import os
from typing import Optional, Union

from labfunctions import defaults, secrets, types
from labfunctions.conf import load_client
from labfunctions.utils import secure_filename

from .diskclient import DiskClient
from .nbclient import NBClient
from .state import WorkflowsState
from .state import from_file as ws_from_file


def from_file(
    filepath="workflows.yaml", url_service=None, home_dir=defaults.CLIENT_HOME_DIR
) -> DiskClient:
    """intialize a py:class:`labfunctions.client.diskclient.DiskClient`
    using data from local like workflows.yaml.
    """

    settings = load_client()
    wf_state = ws_from_file(filepath)
    wf_state._file = filepath
    wf_service = url_service or settings.WORKFLOW_SERVICE
    # if not creds:
    #    creds = login_cli(url_service, home_dir)
    dc = DiskClient(
        url_service=wf_service,
        wf_state=wf_state,
        base_path=settings.BASE_PATH,
    )
    dc.load_creds()
    # if not os.environ.get("DEBUG"):
    #    try:
    #        dc.logincli()
    #    except httpx.ConnectError:
    #        print("No connection with server")
    return dc


def from_env(settings: Optional[types.ClientSettings] = None) -> NBClient:
    """Creates a client using the settings module and environment variables"""
    if not settings:
        settings = load_client()
    nbvars = secrets.load(settings.BASE_PATH)
    at = nbvars.get("NB_ACCESS_TOKEN")
    rt = nbvars.get("NB_REFRESH_TOKEN")
    if at and rt:
        os.environ["NB_ACCESS_TOKEN"] = at
        os.environ["NB_REFRESH_TOKEN"] = rt

    # creds = _load_creds(settings, nbvars)
    pd = types.ProjectData(name=settings.PROJECT_NAME, projectid=settings.PROJECTID)

    wf_state = WorkflowsState(pd)
    c = NBClient(
        url_service=settings.WORKFLOW_SERVICE,
        wf_state=wf_state,
    )
    c.load_creds()
    return c


def agent(url_service, token, refresh, projectid) -> NBClient:
    """A shortcut to intialize

    Usually, it is used in each machine running a agent (data plane),
    and is used to communicates with server for workflows task
    preparations.

    TODO: in the future agent use case should be treat apart. Right now
    if the credential env variable is outdated, then
    it will be allways refreshing their token.

    Workaround #1: extend expire time for services like tokens
    Workaround #2: manage a shared store like
    store_creds_disk with redis instead
    Workaround #3: Inject crendentials with the task.


    :param url_service: WORKFLOWS_SERVICE url
    :param token: access_token
    :param refresh: refresh_token
    :param projectid: projectid
    """
    wf = WorkflowsState(types.ProjectData(name="", projectid=projectid))
    creds = types.TokenCreds(access_token=token, refresh_token=refresh)
    return NBClient(url_service=url_service, creds=creds, wf_state=wf)



from copy import deepcopy
from typing import Any, Dict, List, Optional, Set

import redis
from rq import Queue

from labfunctions.control_plane.register import AgentRegister
from labfunctions.types import ServerSettings
from labfunctions.types.agent import AgentNode
from labfunctions.types.cluster import (
    ClusterDiff,
    ClusterFile,
    ClusterPolicy,
    ClusterSpec,
    ClusterState,
    ScaleIdle,
    ScaleItems,
)
from labfunctions.types.machine import (
    ExecMachineResult,
    ExecutionMachine,
    MachineInstance,
)
from labfunctions.utils import open_yaml

from . import deploy
from .base import ProviderSpec
from .context import machine_from_settings
from .inventory import Inventory


def apply_scale_items(state: ClusterState, scale: ScaleItems) -> ClusterState:
    new_state = deepcopy(state)
    if state.queue_items[scale.qname] >= scale.items_gt:
        new_state.agents_n += scale.increase_by
    elif state.queue_items[scale.qname] <= scale.items_lt:
        new_state.agents_n -= scale.decrease_by
        new_state.agents.pop()

    return new_state


def apply_idle(state: ClusterState, idle: ScaleIdle) -> ClusterState:
    new_state = deepcopy(state)
    shutdown_agents = set()
    for agt in state.agents:
        if state.idle_by_agent[agt] >= idle.idle_time_gt:
            shutdown_agents.add(agt)
    agents = state.agents - shutdown_agents

    new_state.agents = agents
    new_state.agents_n = len(agents)
    return new_state


def apply_minmax(state: ClusterState, policy: ClusterPolicy) -> ClusterState:
    new_state = deepcopy(state)
    if state.agents_n < policy.min_nodes:
        new_state.agents_n = policy.min_nodes
    elif state.agents_n > policy.max_nodes:
        new_state.agents_n = policy.max_nodes

    return new_state


class ClusterControl:

    POLICIES = {
        "items": apply_scale_items,
        "idle": apply_idle,
        # "minmax": apply_minmax
    }

    def __init__(
        self, register: AgentRegister, spec: ClusterSpec, inventory: Inventory
    ):
        self.name = spec.name
        self.spec = spec
        self.inventory = inventory
        self.register = register
        self.state = self.build_state()
        self.provider = self.get_provider()

    @property
    def cluster_name(self):
        return self.name

    def get_provider(self) -> ProviderSpec:
        return self.inventory.get_provider(self.spec.provider)

    def refresh(self):
        self.state = self.build_state()

    def build_state(self) -> ClusterState:
        agents = self.register.list_agents(self.name)
        # agents_obj = {self.register.get(a) for a in agents}
        agents_n = len(agents)
        queues_obj: List[Queue] = [self.register.get_queue(q) for q in self.spec.qnames]
        queue_items = {q.name: len(q) for q in queues_obj}

        idle_by_agent = self.register.idle_agents(agents, queues_obj)

        return ClusterState(
            agents_n=agents_n,
            agents=set(agents),
            queue_items=queue_items,
            idle_by_agent=idle_by_agent,
        )

    @property
    def policy(self) -> ClusterPolicy:
        return self.spec.policy

    def apply_policies(self):
        new_state = self.state.copy()
        for scale in self.policy.strategies:
            new_state = self.POLICIES[scale.name](new_state, scale)

        new_state = apply_minmax(new_state, self.policy)

        return new_state

    def difference(self, new_state: ClusterState) -> ClusterDiff:

        to_delete = self.state.agents - new_state.agents

        to_create = 0
        if new_state.agents_n > self.state.agents_n:
            to_create = new_state.agents_n - self.state.agents_n

        if len(new_state.agents) < self.policy.min_nodes:
            to_create = self.policy.min_nodes - len(new_state.agents)

        return ClusterDiff(to_create=to_create, to_delete=list(to_delete))

    def create_instance(
        self,
        settings: ServerSettings,
        do_deploy=True,
        use_public=False,
        deploy_local=False,
    ) -> MachineInstance:
        ctx = machine_from_settings(
            self.spec.machine,
            cluster=self.name,
            qnames=self.spec.qnames,
            network=self.spec.network,
            location=self.spec.location,
            settings=settings,
            inventory=self.inventory,
        )
        print(f"=> Creating machine: {ctx.machine.name}")
        instance = self.provider.create_machine(ctx.machine)
        ip = instance.private_ips[0]
        if use_public:
            ip = instance.public_ips[0]
        req = deploy.agent_from_settings(
            ip,
            instance.machine_id,
            self.name,
            settings,
            qnames=ctx.qnames,
            worker_procs=ctx.worker_procs,
            docker_version=ctx.docker_version,
        )
        if do_deploy and deploy_local:
            res = deploy.agent_local(req, settings.dict())
            print("=> ", res)
        elif do_deploy:
            res = deploy.agent(req, settings.dict())
            print("=> ", res)
        self.register.register_machine(instance)
        return instance

    def destroy_instance(self, agent_name: str):
        agent = self.register.get(agent_name)
        if agent:
            self.register.kill_workers_from_agent(agent.name)
            machine = self.register.get_machine(agent.machine_id)
            self.register.unregister(agent)
            self.register.unregister_machine(agent.machine_id)
        else:
            machine = agent_name
        self.provider.destroy_machine(machine)

    def scale(
        self,
        settings: ServerSettings,
        use_public=False,
        deploy_local=False,
        do_deploy=True,
    ):
        new_state = self.apply_policies()
        diff = self.difference(new_state)

        print(f"=> To create: {diff.to_create}")
        for _ in range(diff.to_create):
            self.create_instance(
                settings,
                use_public=use_public,
                deploy_local=deploy_local,
                do_deploy=do_deploy,
            )

        for agt in diff.to_delete:
            print(f"=> Deleting agent {agt}")
            self.destroy_instance(agt)



from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator, Dict, Generator, Union

from labfunctions.utils import get_class


class KeyReadError(Exception):
    def __init__(self, bucket, key, error_msg):
        msg = f"Value not found for key {key} in {bucket} with error: {error_msg}"
        super().__init__(msg)


class KeyWriteError(Exception):
    def __init__(self, bucket, key, error_msg):
        msg = f"Value not written for key {key} in {bucket} with error: {error_msg}"
        super().__init__(msg)


class GenericKVSpec(ABC):
    """
    This is a generic KV store mostly use for project data related
    The interface is very simple allowing the common put/get actions
    either loading data in memory or streaming data for big files

    For examples about how to use some of them see tests/test_io_kv.py

    This interface is offered in a sync and async version
    """

    def __init__(self, bucket: str, client_opts: Dict[str, Any] = {}):
        self._opts = client_opts
        self._bucket = bucket

    @abstractmethod
    def put(self, key: str, bdata: bytes):
        pass

    @abstractmethod
    def put_stream(self, key: str, generator: Generator[bytes, None, None]) -> bool:
        pass

    @abstractmethod
    def get(self, key: str) -> Union[bytes, str, None]:
        pass

    @abstractmethod
    def get_stream(self, key: str) -> Generator[bytes, None, None]:
        pass

    @staticmethod
    def create(store_class, bucket, opts: Dict[str, Any] = {}) -> "GenericKVSpec":
        Class = get_class(store_class)
        return Class(bucket, opts)


class AsyncKVSpec(ABC):
    """
    This is a generic KV store mostly use for project data related
    The interface is very simple allowing the common put/get actions
    either loading data in memory or streaming data for big files

    For examples about how to use some of them see tests/test_io_kv.py


    This interface is offered in a sync and async version
    """

    def __init__(self, client_opts: Dict[str, Any], bucket: str):
        self._opts = client_opts
        self._bucket = bucket

    @abstractmethod
    async def put(self, key: str, bdata: bytes):
        pass

    @abstractmethod
    async def put_stream(
        self, key: str, generator: Generator[bytes, None, None]
    ) -> bool:
        pass

    @abstractmethod
    async def get(self, key: str) -> Union[bytes, str, None]:
        pass

    @abstractmethod
    async def get_stream(self, key: str) -> AsyncGenerator[bytes, None]:
        pass

    @staticmethod
    def create(store_class, bucket, opts: Dict[str, Any] = {}) -> "GenericKVSpec":
        Class = get_class(store_class)
        return Class(bucket, opts)



from collections import namedtuple

# Shared defaults by server, agent and client
QueuesNS = namedtuple("QueuesNS", ["control", "machine", "build"])
Q_NS = QueuesNS(control="ctrl", machine="mch", build="bui")
CLOUD_TAG = "nbworkflows"

REFRESH_TOKEN_PATH = "auth/refresh_token"

API_VERSION = "v1"

AGENT_USER_PREFIX = "agt"
AGENT_SCOPES = "agent:rw"
AGENT_ADMIN_SCOPES = "agent:r:w,admin:r"
AGENT_LEN = 8

NOTEBOOKS_DIR = "notebooks/"

# Builder
ZIP_GIT_PREFIX = "src/"

# Secrets and security
SECRETS_FILENAME = ".secrets"
NBVARS_VAR_NAME = "NB_NBVARS"
NBVARS_FILENAME = "local.nbvars"
PRIVKEY_VAR_NAME = "PRIVATE_KEY"

# Client DEFAULT OPTIONS
CLIENT_TMP_FOLDER = ".nb_tmp"
CLIENT_HOME_DIR = ".labfunctions/"
CLIENT_TIMEOUT = 60
CLIENT_CREDS_FILE = "credentials.json"
CLIENT_AGENT_CREDS_FILE = "agent.creds.json"

DOCKERFILE_MAINTENER = "NB Workflows <package@nbworkflows.com>"
DOCKERFILE_IMAGE = "nuxion/labfunctions-client"
DOCKERFILE_IMAGE_GPU = "nuxion/labfunctions-client-gpu"
DOCKER_AUTHOR = "nbworkflows"
DOCKER_UID = "1000"
DOCKER_GID = "997"

# Sanic
SANIC_APP_NAME = "labfunctions"

NB_OUTPUTS = "outputs"

EXECUTIONTASK_VAR = "NB_EXECUTION_TASK"

BASE_PATH_ENV = "NB_BASE_PATH"

PROJECT_UPLOADS = "uploads"
PROJECT_HISTORY = "history"

# see https://zelark.github.io/nano-id-cc/
PROJECT_ID_ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz"
PROJECTID_MIN_LEN = 10  # 13 years 1% collision at 100 projects creations per hour
PROJECTID_MAX_LEN = 16
WFID_LEN = 11  # ~139 thousand years 1% collision at 1000 jobs creation per hour
EXECID_LEN = 14  # ~20 years %1 collision at 1000 execs per second
PIPEID_LEN = 11
NANO_ID_ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz-"
NANO_URLSAFE_ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz-_"
NANO_MACHINE_ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz"

MACHINE_TYPE = "cpu"
CLUSTER_NAME = "default"

AGENT_HOMEDIR = "/home/op"
AGENT_DOCKER_IMG = "nuxion/labfunctions"
AGENT_ENV_TPL = "agent.docker.envfile"

NVIDIA_GPG_VERSION = "2004"
NVIDIA_GPG_KEY = "3bf863cc"



class LoginError(Exception):
    def __init__(self, addr, user):
        msg = f"{user} failed auth to server {addr}"
        super().__init__(msg)


class WorkflowStateNotSetError(Exception):
    def __init__(self, module_name):
        msg = f"WorkflowState is not set in {module_name}"
        super().__init__(msg)


class ProjectUploadError(Exception):
    def __init__(self, projectid):
        msg = f"Upload Failed from project {projectid}"
        super().__init__(msg)


class ProjectCreateError(Exception):
    def __init__(self, projectid):
        msg = f"Project creation error for projectid or name: {projectid}"
        super().__init__(msg)



from typing import Optional

import redis

from labfunctions.control_plane.register import AgentRegister
from labfunctions.errors.cluster import ClusterSpecNotFound
from labfunctions.types.cluster import ClusterSpec

from .cluster_file import load_cluster_file
from .control import ClusterControl
from .inventory import Inventory


def get_spec_from_file(from_file: str, cluster=None) -> ClusterSpec:
    cluster_file = load_cluster_file(from_file)
    try:
        cluster = cluster or cluster_file.default_cluster
        spec = cluster_file.clusters[cluster]
    except KeyError:
        raise ClusterSpecNotFound(cluster, from_file)
    return spec


def create_cluster_control(
    from_file: str, register_url: str, cluster: Optional[str] = None
) -> ClusterControl:
    cluster_file = load_cluster_file(from_file)
    spec = get_spec_from_file(from_file, cluster)

    inventory = Inventory(cluster_file.inventory)
    rdb = redis.from_url(register_url, decode_responses=True)
    registry = AgentRegister(rdb, cluster)
    cc = ClusterControl(registry, spec=spec, inventory=inventory)
    return cc



import signal
import threading
import time

from redis import Redis


class HeartbeatThread(threading.Thread):

    PREFIX = "nb.ag.heart"

    def __init__(self, rdb: Redis, name: str, ttl_secs=15, check_every_secs=10):
        threading.Thread.__init__(self)

        self.rdb = rdb
        self.name = name
        self.ttl_secs = ttl_secs
        self.check_every = check_every_secs
        self.shutdown_flag = threading.Event()

    def run(self):
        key = f"{self.PREFIX}.{self.name}"
        while not self.shutdown_flag.is_set():
            print(f"Heartbeat: I'm alive on key {key}")
            self.rdb.set(key, "alive", ex=self.ttl_secs)
            time.sleep(self.check_every)

    def unregister(self):
        print("Calling unregister")
        key = f"{self.PREFIX}.{self.name}"
        self.rdb.delete(key)
        self.shutdown_flag.set()



from typing import Union

from cryptography.hazmat.primitives.kdf.scrypt import Scrypt


class PasswordScript:
    """
    https://cryptography.io/en/latest/hazmat/primitives/key-derivation-functions/#cryptography.hazmat.primitives.kdf.scrypt.Scrypt
    """

    def __init__(self, salt: Union[bytes, str], n=2**14, r=8, p=1):
        if isinstance(salt, str):
            salt = salt.encode("utf-8")
        self.n = n
        self.r = r
        self.p = p
        self.salt = salt

    def _kdf(self):
        kdf = Scrypt(
            salt=self.salt,
            length=32,
            n=self.n,
            r=self.r,
            p=self.p,
        )
        return kdf

    def encrypt(self, pass_: str) -> bytes:
        kdf = self._kdf()

        return kdf.derive(pass_.encode("utf-8"))

    def verify(self, unencypted: str, encrypted: bytes) -> bool:
        """Verify if a uncrypted text match with an encrypted one"""
        kdf = self._kdf()
        try:
            kdf.verify(unencypted.encode("utf-8"), encrypted)
            return True
        except:
            return False



import os
import sys
import time
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from libcloud.compute.base import Node, NodeLocation
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider
from pydantic import BaseSettings

from labfunctions import defaults
from labfunctions.hashes import generate_random
from labfunctions.types.machine import (
    BlockInstance,
    BlockStorage,
    ExecMachineResult,
    ExecutionMachine,
    MachineInstance,
    MachineRequest,
)

from .base import ProviderSpec


class GCConf(BaseSettings):
    service_account: str
    project: str
    pem_file: Optional[str] = None
    datacenter: Optional[str] = None
    credential_file: Optional[str] = None

    class Config:
        env_prefix = "NB_GCE_"


def get_gce_driver():
    GCE = get_driver(Provider.GCE)
    return GCE


def generic_zone(name, driver) -> NodeLocation:
    return NodeLocation(id=None, name=name, country=None, driver=driver)


class GCEProvider(ProviderSpec):
    def __init__(self, conf: Optional[GCConf] = None):
        self.conf = conf or GCConf()
        G = get_gce_driver()
        # _env_creds = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
        # if _env_creds:
        #     conf.credential_file = _env_creds

        self.driver = G(
            self.conf.service_account,
            key=self.conf.pem_file,
            project=self.conf.project,
            datacenter=self.conf.datacenter,
        )

    def _get_volume(self, vol_name):
        volumes = [v for v in self.driver.list_volumes() if v.name == vol_name]
        if len(volumes) > 0:
            return volumes[0]
        return None

    def _get_volumes_to_attach(self, volumes: List[BlockStorage]):
        to_attach = []
        for vol in volumes:
            v = self._get_volume(vol.name)
            if v:
                to_attach.append(v)
            if not v and vol.create_if_not_exist:
                created = self.driver.create_volume(
                    vol.size,
                    vol.name,
                    location=vol.location,
                    snapshot=vol.snapshot,
                    ex_disk_type=vol.kind,
                )
                to_attach.append(created)
        if to_attach:
            return to_attach
        return None

    def _create_boot_disk(
        self, volumes: List[BlockStorage]
    ) -> Union[BlockInstance, None]:
        instance = None
        for v in volumes:
            if v.mount == "/":
                random_name = generate_random(
                    size=6, alphabet=defaults.NANO_MACHINE_ALPHABET
                )
                v.name = f"boot-{random_name}"
                instance = self.create_volume(v)
                break
        return instance

    def create_machine(self, node: MachineRequest) -> MachineInstance:
        metadata = {
            "items": [
                {"key": "ssh-keys", "value": f"{node.ssh_user}: {node.ssh_public_cert}"}
            ]
        }
        volumes = None
        if node.volumes:
            non_boot = [v for v in node.volumes if v.mount != "/"]
            volumes = self._get_volumes_to_attach(non_boot)

        boot = self._create_boot_disk(node.volumes)
        boot_name = None
        if boot:
            boot_name = boot.name

        tags = node.labels.get("tags")
        _labels = deepcopy(node.labels)
        del _labels["tags"]
        labels = _labels

        maintence_policy = None
        accelerator_type = None
        accelerator_count = None
        if node.gpu:
            maintence_policy = "TERMINATE"
            accelerator_type = node.gpu.gpu_type
            accelerator_count = node.gpu.count

        instance = self.driver.create_node(
            node.name,
            size=node.size,
            image=node.image,
            location=node.location,
            ex_boot_disk=boot_name,
            ex_network=node.network,
            ex_metadata=metadata,
            ex_tags=tags,
            ex_labels=labels,
            ex_accelerator_type=accelerator_type,
            ex_accelerator_count=accelerator_count,
            ex_on_host_maintenance=maintence_policy,
        )

        attached = []
        if volumes:
            for v in volumes:
                _attached = self.driver.attach_volume(instance, v)
                if _attached:
                    attached.append(v.name)
            if boot:
                attached.append(boot.name)

        res = MachineInstance(
            machine_id=f"/gce/{node.location}/{node.name}",
            machine_name=node.name,
            location=node.location,
            volumes=attached,
            main_addr=instance.private_ips[0],
            private_ips=instance.private_ips,
            public_ips=instance.public_ips,
        )
        return res

    def list_machines(
        self, location: Optional[str] = None, tags: Optional[List[str]] = None
    ) -> List[MachineInstance]:
        nodes = self.driver.list_nodes(ex_zone=location)
        filtered_nodes = []
        if tags:
            for n in nodes:
                if n.state == "running" and n.extra["tags"]:
                    for t in tags:
                        if t in n.extra["tags"]:
                            filtered_nodes.append(n)
        else:
            filtered_nodes = [n for n in nodes if n.state == "running"]
        final = []
        for n in filtered_nodes:
            _lbl = n.extra["labels"] if n.extra["labels"] else {}
            labels = {"tags": n.extra["tags"], **_lbl}
            _n = MachineInstance(
                machine_id=n.id,
                machine_name=n.name,
                location=n.extra["zone"].name,
                labels=labels,
                main_addr=n.private_ips[0],
                private_ips=n.private_ips,
                public_ips=n.public_ips,
            )
            final.append(_n)
        return final

    def destroy_machine(self, node: Union[str, MachineInstance]):
        name = node
        if isinstance(node, MachineInstance):
            name = node.machine_name
        nodes = self.driver.list_nodes()
        _node = [n for n in nodes if n.name == name][0]
        _node.destroy()

    def create_volume(self, disk: BlockStorage) -> BlockInstance:
        vol = self.driver.create_volume(
            disk.size,
            disk.name,
            location=disk.location,
            image=disk.image,
            snapshot=disk.snapshot,
            ex_disk_type=disk.kind,
        )
        block = BlockInstance(id=vol.id, **disk.dict())
        block.extra = vol.extra
        return block

    def destroy_volume(self, disk: Union[str, BlockStorage]) -> bool:
        name = disk
        if isinstance(disk, BlockStorage):
            name = disk.name
        vol = [v for v in self.driver.list_volumes() if v.name == name][0]
        rsp = self.driver.destroy_volume(vol)
        return rsp

    def attach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        _node = [n for n in self.driver.list_nodes() if n.name == node.machine_name][0]
        vol = [v for v in self.driver.list_volumes() if v.name == disk.name]

        res = self.driver.attach_volume(_node, vol)
        return res

    def detach_volume(self, node: MachineInstance, disk: BlockStorage) -> bool:
        _node = [n for n in self.driver.list_nodes() if n.name == node.machine_name][0]
        vol = [v for v in self.driver.list_volumes() if v.name == disk.name]
        res = self.driver.detach_volume(vol, _node)
        return res



from typing import List, Optional, Union

from sqlalchemy import delete as sqldelete
from sqlalchemy import insert as sqlinsert
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload

from labfunctions.errors.runtimes import RuntimeNotFound
from labfunctions.models import RuntimeModel
from labfunctions.types.runtimes import RuntimeData, RuntimeReq, RuntimeSpec
from labfunctions.utils import get_version


def select_runtime():
    stmt = select(RuntimeModel).options(selectinload(RuntimeModel.project))
    return stmt


def model2runtime(m: RuntimeModel) -> RuntimeData:
    rd = RuntimeData(
        id=m.id,
        runtimeid=m.runtimeid,
        runtime_name=m.runtime_name,
        docker_name=m.docker_name,
        spec=RuntimeSpec(**m.spec),
        project_id=m.project.projectid,
        registry=m.registry,
        created_at=m.created_at.isoformat(),
        version=m.version,
    )
    return rd


def _insert(rd: RuntimeData):
    t = RuntimeModel.__table__
    stmt = sqlinsert(t).values(rd.dict(exclude_none=True))
    return stmt


async def get_list(session, projectid, limit=10) -> List[RuntimeData]:
    stmt = (
        select_runtime()
        .where(RuntimeModel.project_id == projectid)
        .order_by(RuntimeModel.created_at.desc())
        .limit(limit)
    )
    rows = await session.execute(stmt)

    return [model2runtime(r[0]) for r in rows]


async def create(session, rq: RuntimeReq) -> bool:
    rd = RuntimeData(
        runtimeid=f"{rq.project_id}/{rq.runtime_name}/{rq.version}", **rq.dict()
    )
    stmt = _insert(rd)
    inserted = True
    try:
        await session.execute(stmt)
    except IntegrityError as e:
        inserted = False
    return inserted


async def get_by_rid(session, runtimeid: str) -> Union[RuntimeData, None]:
    stmt = select_runtime().where(RuntimeModel.runtimeid == runtimeid).limit(1)

    rsp = await session.execute(stmt)
    model = rsp.scalar_one_or_none()
    if model:
        rd = model2runtime(model)
        return rd
    return None


async def get_runtime(
    session, projectid: str, runtime_name: str, version=None
) -> Union[RuntimeData, None]:
    stmt = select_runtime().where(RuntimeModel.project_id == projectid)
    if version:
        stmt = (
            stmt.where(RuntimeModel.runtime_name == runtime_name)
            .where(RuntimeModel.version == version)
            .limit(1)
        )
    else:
        stmt = (
            stmt.where(RuntimeModel.runtime_name == runtime_name)
            .order_by(RuntimeModel.created_at.desc())
            .limit(1)
        )

    rsp = await session.execute(stmt)
    model = rsp.scalar_one_or_none()
    if model:
        rd = model2runtime(model)
        return rd
    return None


def get_by_rid_sync(session, runtimeid: str) -> Union[RuntimeData, None]:
    stmt = select_runtime().where(RuntimeModel.runtimeid == runtimeid).limit(1)

    rsp = session.execute(stmt)
    model = rsp.scalar_one_or_none()
    if model:
        rd = model2runtime(model)
        return rd
    return None


async def delete_by_rid(session, runtimeid: int):
    stmt = sqldelete(RuntimeModel).where(RuntimeModel.runtimeid == runtimeid)
    await session.execute(stmt)


# def docker_name_from_runtime(session, runtimeid: Optional[str] = None) -> RuntimeData:
#     """ It returns the final docker name based on a runtime id,
#     if an id is not provided then it will returns the default runtime. """
#     if runtime:
#         _runtime = get_by_rid_sync(session, runtime)
#         if not _runtime:
#             raise RuntimeNotFound(runtime)
#         runtime = f"{_runtime.docker_name}:{_runtime.version}"
#         if _runtime.registry:
#             runtime = f"{_runtime.registry}/{runtime}"
#     else:
#         version = get_version()
#         runtime = f"nuxion/labfunctions:{version}"
#     return runtime
